"""
Basic reaggregation

This is called 'basic' because it's the first one we thought about.
It's also, in some ways, the simplest.
It assumes that domestic aviation is reported at the model region level.
"""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Callable

import numpy as np
import pandas as pd
from attrs import define, field
from pandas_openscm.grouping import groupby_except
from pandas_openscm.index_manipulation import (
    set_levels,
    update_index_levels_func,
)
from pandas_openscm.indexing import multi_index_lookup

from gcages.aggregation import aggregate_df_level, get_region_sector_sum
from gcages.assertions import assert_only_working_on_variable_unit_region_variations
from gcages.cmip7_scenariomip.gridding_emissions import (
    COMPLETE_GRIDDING_SPECIES,
    SpatialResolutionOption,
)
from gcages.cmip7_scenariomip.pre_processing.reaggregation.common import (
    ToCompleteResult,
)
from gcages.completeness import assert_all_groups_are_complete, get_missing_levels
from gcages.index_manipulation import (
    combine_sectors,
    create_levels_based_on_existing,
    set_new_single_value_levels,
    split_sectors,
)
from gcages.internal_consistency import InternalConsistencyError
from gcages.testing import compare_close, get_variable_unit_default
from gcages.typing import NP_ARRAY_OF_FLOAT_OR_INT
from gcages.units_helpers import convert_unit_like

if TYPE_CHECKING:
    from gcages.typing import PINT_SCALAR


@define
class GriddingSectorComponents:
    """
    Definition of the components of a gridding sector for reporting

    This is meant for internal use only.

    OR logic is applied to the exclusions
    i.e. a variable will not be required
    if the sector is in `input_sectors_optional`
    or the species is in `input_species_optional`
    (i.e. we are maximally relaxed about optional reporting,
    instead of using AND logic and being restrictive).
    """

    gridding_sector: str
    """The gridding sector"""

    spatial_resolution: SpatialResolutionOption

    input_sectors: tuple[str, ...]
    """
    The input sectors
    """

    input_sectors_optional: tuple[str, ...]
    """
    The input sectors that are optional
    """

    all_species: tuple[str, ...]
    """The input species"""

    input_species_optional: tuple[str, ...]
    """The input species that are optional"""

    reporting_only: bool
    """Is this definition only used for reporting, not aggregating?"""

    def to_complete_variables(self) -> tuple[str, ...]:
        """
        Convert to the complete set of variables for this gridding sector
        """
        return tuple(
            f"Emissions|{species}|{sector}"
            for species in self.all_species
            for sector in self.input_sectors
        )

    def to_required_variables(self) -> tuple[str, ...]:
        """
        Convert to the required set of variables for this gridding sector
        """
        return tuple(
            f"Emissions|{species}|{sector}"
            for species in self.all_species
            for sector in self.input_sectors
            if not (
                sector in self.input_sectors_optional
                or species in self.input_species_optional
            )
        )


@define
class GriddingSectorComponentsCarbonRemovalReporting:
    """
    Definition of the components of a carbon removal gridding sector for reporting

    This is for carbon removal i.e. is for CO2 only.

    This is meant for internal use only.
    """

    gridding_sector: str
    """The gridding sector"""

    spatial_resolution: SpatialResolutionOption

    input_sectors: tuple[str, ...]
    """
    The input sectors
    """

    input_sectors_optional: tuple[str, ...]
    """
    The input sectors that are optional
    """

    reporting_only: bool
    """Is this definition only used for reporting, not aggregating?"""

    def to_complete_variables(self) -> tuple[str, ...]:
        """
        Convert to the complete set of variables for this gridding sector
        """
        return tuple(f"Carbon Removal|{sector}" for sector in self.input_sectors)

    def to_required_variables(self) -> tuple[str, ...]:
        """
        Convert to the required set of variables for this gridding sector
        """
        return tuple(
            f"Carbon Removal|{sector}"
            for sector in self.input_sectors
            if sector not in self.input_sectors_optional
        )


SECTOR_DOMESTIC_AVIATION = "Energy|Demand|Transportation|Domestic Aviation"
"""
Domestic aviation sector
"""

gridding_sectors_reporting = (
    GriddingSectorComponents(
        gridding_sector="Agricultural Waste Burning",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("AFOLU|Agricultural Waste Burning",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=("CO2",),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Agriculture",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=(
            "AFOLU|Agriculture",
            "AFOLU|Land|Harvested Wood Products",
            "AFOLU|Land|Land Use and Land-Use Change",
            "AFOLU|Land|Other",
            "AFOLU|Land|Wetlands",
        ),
        input_sectors_optional=(
            "AFOLU|Land|Harvested Wood Products",
            "AFOLU|Land|Land Use and Land-Use Change",
            "AFOLU|Land|Other",
            "AFOLU|Land|Wetlands",
        ),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=(
            "BC",
            "CO",
            "OC",
            "CO2",
            "Sulfur",
        ),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Aircraft",
        spatial_resolution=SpatialResolutionOption.WORLD,
        input_sectors=(
            "Energy|Demand|Bunkers|International Aviation",
            # Domestic aviation is included too.
            # However, it has to be reported at the regional level
            # so we can subtract it from Transport
            # (hence it doesn't appear here, see below)
        ),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=("CH4",),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Domestic aviation headache",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=(SECTOR_DOMESTIC_AVIATION,),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=("CH4",),
        reporting_only=True,
    ),
    GriddingSectorComponents(
        gridding_sector="Transportation Sector",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Energy|Demand|Transportation",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=(),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Energy Sector",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Energy|Supply",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=(),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Forest Burning",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("AFOLU|Land|Fires|Forest Burning",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=("CO2",),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Grassland Burning",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("AFOLU|Land|Fires|Grassland Burning",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=("CO2",),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Industrial Sector",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=(
            "Energy|Demand|Industry",
            "Energy|Demand|Other Sector",
            "Industrial Processes",
            "Other",
        ),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_sectors_optional=(
            "Energy|Demand|Other Sector",
            "Other",
        ),
        input_species_optional=(),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="International Shipping",
        spatial_resolution=SpatialResolutionOption.WORLD,
        input_sectors=("Energy|Demand|Bunkers|International Shipping",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=(),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Peat Burning",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("AFOLU|Land|Fires|Peat Burning",),
        input_sectors_optional=("AFOLU|Land|Fires|Peat Burning",),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=COMPLETE_GRIDDING_SPECIES,
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Residential Commercial Other",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Energy|Demand|Residential and Commercial and AFOFI",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=(),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Solvents Production and Application",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Product Use",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=("BC", "CH4", "CO", "NOx", "OC", "Sulfur"),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Waste",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Waste",),
        input_sectors_optional=(),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=(),
        reporting_only=False,
    ),
    GriddingSectorComponents(
        gridding_sector="Other CDR",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Other Capture and Removal",),
        input_sectors_optional=("Other Capture and Removal",),
        all_species=COMPLETE_GRIDDING_SPECIES,
        input_species_optional=(
            "CH4",
            "N2O",
            "BC",
            "CO",
            "NH3",
            "OC",
            "NOx",
            "Sulfur",
            "VOC",
        ),
        reporting_only=False,
    ),
    GriddingSectorComponentsCarbonRemovalReporting(
        gridding_sector="BECCS",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Geological Storage|Biomass",),
        input_sectors_optional=("Geological Storage|Biomass",),
        reporting_only=False,
    ),
    GriddingSectorComponentsCarbonRemovalReporting(
        gridding_sector="Enhanced Weathering",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Enhanced Weathering",),
        input_sectors_optional=("Enhanced Weathering",),
        reporting_only=False,
    ),
    GriddingSectorComponentsCarbonRemovalReporting(
        gridding_sector="Direct Air Capture",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Geological Storage|Direct Air Capture",),
        input_sectors_optional=("Geological Storage|Direct Air Capture",),
        reporting_only=False,
    ),
    GriddingSectorComponentsCarbonRemovalReporting(
        gridding_sector="Ocean",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Ocean",),
        input_sectors_optional=("Ocean",),
        reporting_only=False,
    ),
    GriddingSectorComponentsCarbonRemovalReporting(
        gridding_sector="Biochar",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Land Use|Biochar",),
        input_sectors_optional=("Land Use|Biochar",),
        reporting_only=False,
    ),
    GriddingSectorComponentsCarbonRemovalReporting(
        gridding_sector="Soil Carbon Management",
        spatial_resolution=SpatialResolutionOption.MODEL_REGION,
        input_sectors=("Land Use|Soil Carbon Management",),
        input_sectors_optional=("Land Use|Soil Carbon Management",),
        reporting_only=False,
    ),
)
"""
The reporting sector component definitions

Note that this only defines the reporting sectors.
The sectors used for aggregation are defined in [to_gridding_sectors][(m).]
because the logic for how we do the re-aggregation is more complex
than a straight mapping.
"""

COMPLETE_WORLD_VARIABLES: tuple[str, ...] = tuple(
    v
    for gs in gridding_sectors_reporting
    if gs.spatial_resolution == SpatialResolutionOption.WORLD
    for v in gs.to_complete_variables()
)
"""
Complete set of variables at the world level
"""

REQUIRED_WORLD_VARIABLES: tuple[str, ...] = tuple(
    v
    for gs in gridding_sectors_reporting
    if gs.spatial_resolution == SpatialResolutionOption.WORLD
    for v in gs.to_required_variables()
)
"""
Required set of variables at the world level
"""

OPTIONAL_WORLD_VARIABLES: tuple[str, ...] = tuple(
    set(COMPLETE_WORLD_VARIABLES) - set(REQUIRED_WORLD_VARIABLES)
)
"""
Optional set of variables at the world level
"""

COMPLETE_MODEL_REGION_VARIABLES: tuple[str, ...] = tuple(
    v
    for gs in gridding_sectors_reporting
    if gs.spatial_resolution == SpatialResolutionOption.MODEL_REGION
    for v in gs.to_complete_variables()
)
"""
Complete set of variables at the model region level
"""

REQUIRED_MODEL_REGION_VARIABLES: tuple[str, ...] = tuple(
    v
    for gs in gridding_sectors_reporting
    if gs.spatial_resolution == SpatialResolutionOption.MODEL_REGION
    for v in gs.to_required_variables()
)
"""
Required set of variables at the model region level
"""

OPTIONAL_MODEL_REGION_VARIABLES: tuple[str, ...] = tuple(
    set(COMPLETE_MODEL_REGION_VARIABLES) - set(REQUIRED_MODEL_REGION_VARIABLES)
)
"""
Optional set of variables at the model region level
"""


def get_complete_timeseries_index(
    model_regions: tuple[str, ...],
    world_region: str = "World",
    region_level: str = "region",
    variable_level: str = "variable",
) -> pd.MultiIndex:
    """
    Get the index of complete data

    Parameters
    ----------
    model_regions
        Model regions to use while reaggregating

    world_region
        The value used when the data represents the sum over all regions

    region_level
        Region level in the data index

    variable_level
        Variable level in the data index

    Returns
    -------
    :
        Index of complete data
    """
    world_required = pd.MultiIndex.from_product(
        [COMPLETE_WORLD_VARIABLES, [world_region]], names=[variable_level, region_level]
    )

    model_region_required = pd.MultiIndex.from_product(
        [COMPLETE_MODEL_REGION_VARIABLES, model_regions],
        names=[variable_level, region_level],
    )

    res: pd.MultiIndex = world_required.append(model_region_required)  # type: ignore

    return res


def get_required_timeseries_index(
    model_regions: tuple[str, ...],
    world_region: str = "World",
    region_level: str = "region",
    variable_level: str = "variable",
) -> pd.MultiIndex:
    """
    Get the index of required data

    Parameters
    ----------
    model_regions
        Model regions to use while reaggregating

    world_region
        The value used when the data represents the sum over all regions

    region_level
        Region level in the data index

    variable_level
        Variable level in the data index

    Returns
    -------
    :
        Index for required data
    """
    world_required = pd.MultiIndex.from_product(
        [REQUIRED_WORLD_VARIABLES, [world_region]], names=[variable_level, region_level]
    )

    model_region_required = pd.MultiIndex.from_product(
        [REQUIRED_MODEL_REGION_VARIABLES, model_regions],
        names=[variable_level, region_level],
    )

    res: pd.MultiIndex = world_required.append(model_region_required)  # type: ignore

    return res


def get_internal_consistency_checking_index(
    model_regions: tuple[str, ...],
    world_region: str = "World",
    region_level: str = "region",
    variable_level: str = "variable",
) -> pd.MultiIndex:
    """
    Get the index which selects only data relevant for checking internal consistency

    Parameters
    ----------
    model_regions
        Model regions to use while reaggregating

    world_region
        The value used when the data represents the sum over all regions

    region_level
        Region level in the data index

    variable_level
        Variable level in the data index

    Returns
    -------
    :
        Internal consistency checking index
    """
    world_internal_consistency_checking = pd.MultiIndex.from_product(
        [COMPLETE_WORLD_VARIABLES, [world_region]], names=[variable_level, region_level]
    )
    model_region_consistency_checking_variables = [
        v
        for v in COMPLETE_MODEL_REGION_VARIABLES
        if not (
            # Avoid double counting with "Energy|Demand|Transportation"
            SECTOR_DOMESTIC_AVIATION in v
            or (v.startswith("Carbon Removal") and v.count("|") <= 1)
        )
    ]
    # Add components of the tree that are needed for consistency checking
    # but we don't otherwise use.
    # This is a nasty hack that we should clean up in future
    # so the logic of this is more obvious.
    model_region_consistency_checking_variables.append(
        "Carbon Removal|Land Use|Agroforestry"
    )
    model_region_consistency_checking_variables.append(
        "Carbon Removal|Land Use|Re/Afforestation"
    )
    model_region_consistency_checking_variables.append(
        "Carbon Removal|Land Use|Forest Management"
    )
    model_region_consistency_checking_variables.append("Carbon Removal|Land Use|Other")
    model_region_consistency_checking_variables.append(
        "Carbon Removal|Geological Storage|Other Sources"
    )
    model_region_consistency_checking_variables.append(
        "Carbon Removal|Geological Storage|Synthetic Fuels"
    )
    model_region_consistency_checking = pd.MultiIndex.from_product(
        [model_region_consistency_checking_variables, model_regions],
        names=[variable_level, region_level],
    )

    res: pd.MultiIndex = world_internal_consistency_checking.append(  # type: ignore
        model_region_consistency_checking
    )

    return res


def get_example_input(  # noqa: PLR0913
    model_regions: tuple[str, ...],
    global_only_variables: tuple[tuple[str, str], ...] = (
        ("Emissions|HFC|HFC23", "kt HFC23/yr"),
        ("Emissions|HFC", "kt HFC134a-equiv/yr"),
        ("Emissions|HFC|HFC134a", "kt HFC134a/yr"),
        ("Emissions|HFC|HFC43-10", "kt HFC43-10/yr"),
        ("Emissions|PFC", "kt CF4-equiv/yr"),
        ("Emissions|F-Gases", "Mt CO2-equiv/yr"),
        ("Emissions|SF6", "kt SF6/yr"),
        ("Emissions|CF4", "kt CF4/yr"),
        ("Emissions|C2F6", "kt C2F6/yr"),
        ("Emissions|C6F14", "kt C6F14/yr"),
    ),
    timepoints: NP_ARRAY_OF_FLOAT_OR_INT = np.arange(2010, 2100 + 1, 10.0),
    get_variable_unit: Callable[[str], str] = get_variable_unit_default,
    rng: np.random.Generator = np.random.default_rng(),
    world_region: str = "World",
    model: str = "model",
    scenario: str = "scenario",
    region_level: str = "region",
    variable_level: str = "variable",
    model_level: str = "model",
    scenario_level: str = "scenario",
    unit_level: str = "unit",
    columns_name: str = "year",
) -> pd.DataFrame:
    """
    Get example input data

    Parameters
    ----------
    model_regions
        Model regions to use in the example data

    global_only_variables
        Variables to include only at the global total level

    timepoints
        Timepoints to use in the example

    get_variable_unit
        Function to use to get the unit for each variable

    rng
        Random number generator

    world_region
        The value used when the data represents the sum over all regions

    model
        Model metadata value

    scenario
        Scenario metadata value

    region_level
        Region level in the data index

    variable_level
        Variable level in the data index

    model_level
        Model level in the data index

    scenario_level
        Scenario level in the data index

    unit_level
        Unit level in the data index

    columns_name
        Name of the columns in the output

    Returns
    -------
    :
        Example input expected by the re-aggregator
    """
    # Hard-coded because the example input
    # is tightly coupled to the idea of completeness
    starting_variables = (*COMPLETE_MODEL_REGION_VARIABLES, *COMPLETE_WORLD_VARIABLES)

    # Get our starting point
    start_index = pd.MultiIndex.from_product(
        [starting_variables, model_regions, [model], [scenario]],
        names=[variable_level, region_level, model_level, scenario_level],
    )
    start = pd.DataFrame(
        rng.random((start_index.shape[0], timepoints.size)),
        columns=pd.Index(timepoints, name=columns_name),
        index=start_index,
    )

    # Aggregate up the sectors
    start_sector_full = aggregate_df_level(
        start, level="variable", on_clash="overwrite"
    )
    # Aggregate up the regions
    start_sector_full_region_sum = set_new_single_value_levels(
        groupby_except(start_sector_full, region_level).sum(),
        {region_level: world_region},
    )

    # Put it altogether
    all_info = pd.concat(
        [
            start_sector_full,
            start_sector_full_region_sum.reorder_levels(start_sector_full.index.names),
        ]
    )

    # Keep only the bits required for completeness
    complete_index = get_complete_timeseries_index(
        model_regions=model_regions,
        world_region=world_region,
        region_level=region_level,
        variable_level=variable_level,
    )
    res_gridding = multi_index_lookup(all_info, complete_index)

    # Add unit info
    res_gridding.index = create_levels_based_on_existing(
        res_gridding.index,  # type: ignore # fix when moving to pandas-openscm
        {unit_level: (variable_level, get_variable_unit)},
    )

    if global_only_variables:
        global_only = pd.DataFrame(
            rng.random((len(global_only_variables), timepoints.size)),
            columns=res_gridding.columns,
            index=pd.MultiIndex.from_tuples(
                [
                    (variable, unit, world_region, model, scenario)
                    for (variable, unit) in global_only_variables
                ],
                names=[
                    variable_level,
                    unit_level,
                    region_level,
                    model_level,
                    scenario_level,
                ],
            ),
        )

        res = pd.concat(
            [res_gridding, global_only.reorder_levels(res_gridding.index.names)]
        )

    else:
        res = res_gridding

    return res


def assert_has_all_required_timeseries(
    df: pd.DataFrame,
    model_regions: tuple[str, ...],
    world_region: str = "World",
    region_level: str = "region",
    variable_level: str = "variable",
) -> None:
    """
    Assert that the data has all the required timeseries

    Parameters
    ----------
    df
        Data to check

    model_regions
        Model regions to use while reaggregating

    world_region
        The value used when the data represents the sum over all regions

    region_level
        Region level in the data index

    variable_level
        Variable level in the data index

    Raises
    ------
    NotCompleteError
        `indf` is not complete
    """
    assert_all_groups_are_complete(
        df,
        complete_index=get_required_timeseries_index(
            model_regions=model_regions,
            world_region=world_region,
            region_level=region_level,
            variable_level=variable_level,
        ),
    )


def get_default_internal_conistency_checking_tolerances() -> (
    Mapping[str, Mapping[str, float]] | Mapping[str, Mapping[str, PINT_SCALAR]]
):
    """
    Get default tolerances used when checking the internal consistency of data

    If [openscm_units][] is available,
    we use [pint](https://pint.readthedocs.io)
    quantities for the tolerances to add unit awareness.
    If not, we return plain floats.

    Returns
    -------
    :
        Tolerances to use when checking the internal consistency of the data
    """
    try:
        import openscm_units

        Q = openscm_units.unit_registry.Quantity

        default_tolerances: (
            Mapping[str, Mapping[str, float]] | Mapping[str, Mapping[str, PINT_SCALAR]]
        ) = {  # type: ignore # some issue with openscm-units type hints
            "Emissions|BC": dict(rtol=1e-3, atol=Q(1e-3, "Mt BC/yr")),
            "Emissions|CH4": dict(rtol=1e-3, atol=Q(1e-2, "Mt CH4/yr")),
            "Emissions|CO": dict(rtol=1e-3, atol=Q(1e-1, "Mt CO/yr")),
            "Emissions|CO2": dict(rtol=1e-3, atol=Q(1e0, "Mt CO2/yr")),
            "Emissions|NH3": dict(rtol=1e-3, atol=Q(1e-2, "Mt NH3/yr")),
            "Emissions|NOx": dict(rtol=1e-3, atol=Q(1e-2, "Mt NO2/yr")),
            "Emissions|OC": dict(rtol=1e-3, atol=Q(1e-3, "Mt OC/yr")),
            "Emissions|Sulfur": dict(rtol=1e-3, atol=Q(1e-2, "Mt SO2/yr")),
            "Emissions|VOC": dict(rtol=1e-3, atol=Q(1e-2, "Mt VOC/yr")),
            "Emissions|N2O": dict(rtol=1e-3, atol=Q(1e-1, "kt N2O/yr")),
            "Carbon Removal|Enhanced Weathering": dict(
                rtol=1e-3, atol=Q(1e0, "Mt CO2/yr")
            ),
            "Carbon Removal|Geological Storage": dict(
                rtol=1e-3, atol=Q(1e0, "Mt CO2/yr")
            ),
            "Carbon Removal|Long-Lived Materials": dict(
                rtol=1e-3, atol=Q(1e0, "Mt CO2/yr")
            ),
            "Carbon Removal|Ocean": dict(rtol=1e-3, atol=Q(1e0, "Mt CO2/yr")),
            "Carbon Removal|Land Use": dict(rtol=1e-3, atol=Q(1e0, "Mt CO2/yr")),
            "Carbon Removal|Other": dict(rtol=1e-3, atol=Q(1e0, "Mt CO2/yr")),
        }

    except ImportError:
        default_tolerances = {
            "Emissions|BC": dict(rtol=1e-3, atol=1e-6),
            "Emissions|CH4": dict(rtol=1e-3, atol=1e-6),
            "Emissions|CO": dict(rtol=1e-3, atol=1e-6),
            "Emissions|CO2": dict(rtol=1e-3, atol=1e-6),
            "Emissions|NH3": dict(rtol=1e-3, atol=1e-6),
            "Emissions|NOx": dict(rtol=1e-3, atol=1e-6),
            "Emissions|OC": dict(rtol=1e-3, atol=1e-6),
            "Emissions|Sulfur": dict(rtol=1e-3, atol=1e-6),
            "Emissions|VOC": dict(rtol=1e-3, atol=1e-6),
            "Emissions|N2O": dict(rtol=1e-3, atol=1e-6),
            "Carbon Removal|Enhanced Weathering": dict(rtol=1e-3, atol=1e-6),
            "Carbon Removal|Geological Storage": dict(rtol=1e-3, atol=1e-6),
            "Carbon Removal|Long-Lived Materials": dict(rtol=1e-3, atol=1e-6),
            "Carbon Removal|Ocean": dict(rtol=1e-3, atol=1e-6),
            "Carbon Removal|Land Use|Biochar": dict(rtol=1e-3, atol=1e-6),
            "Carbon Removal|Land Use|Soil Carbon Management": dict(
                rtol=1e-3, atol=1e-6
            ),
            "Carbon Removal|Other": dict(rtol=1e-3, atol=1e-6),
        }

    return default_tolerances


def assert_is_internally_consistent(  # noqa: PLR0913
    df: pd.DataFrame,
    model_regions: tuple[str, ...],
    tolerances: Mapping[str, Mapping[str, float]]
    | Mapping[str, Mapping[str, PINT_SCALAR]],
    world_region: str = "World",
    region_level: str = "region",
    unit_level: str = "unit",
    variable_level: str = "variable",
) -> None:
    """
    Assert that the data is internally consistent

    Parameters
    ----------
    df
        Data to check

    model_regions
        Model regions to use while reaggregating

    tolerances
        Tolerances to apply while checking internal consistency

        Each key should be a variable up to species info
        e.g. "Emission|CH4"
        and each value should be the tolerance arguments to pass
        to [np.isclose][numpy.isclose].
        These tolerance arguments can be pint quantities,
        in which case they are converted to the data's units
        before passing to [np.isclose][numpy.isclose].

    world_region
        The value used when the data represents the sum over all regions

    region_level
        Region level in the data index

    unit_level
        Unit level in the data index

    variable_level
        Variable level in the data index

    Raises
    ------
    InternalConsistencyError
        The data is not internally consistent at the given tolerances
    """
    try:
        import pint
    except ModuleNotFoundError:
        pint = None  # type: ignore

    internal_consistency_checking_index = get_internal_consistency_checking_index(
        model_regions=model_regions,
        world_region=world_region,
        region_level=region_level,
        variable_level=variable_level,
    )

    # Hard-code the logic here
    # because that's what is needed for consistency with the rest of the module.
    # This is one of the issues with the data model used by ScmRun, pyam etc.:
    # the convention is that the first two levels are the species total
    # but this is only implied and not explicit at all
    # (and not even followed in all cases e.g. Emissions|HFC|HFC23).
    def get_species_total_variable(v: str) -> str:
        return "|".join(v.split("|")[:2])

    for species_total_variable, df_species in df.groupby(
        df.index.get_level_values(variable_level).map(get_species_total_variable)
    ):
        df_species_total_reported = df_species.loc[
            (
                df_species.index.get_level_values(variable_level)
                == species_total_variable
            )
            & (df_species.index.get_level_values(region_level) == world_region)
        ]
        if df_species_total_reported.empty:
            # Nothing reported so can move on
            continue

        df_species_internal_consistency_checking_relevant = multi_index_lookup(
            df_species, internal_consistency_checking_index
        )
        if df_species_internal_consistency_checking_relevant.empty:
            # Nothing relevant for checking internal consistency
            continue

        # Note: what we're checking here is that if you sum over the sectors
        # and regions using the index which provides internal consistency,
        # you should get the reported totals.
        # This is different to checking that the implied hierarchy
        # from the variable names
        # (i.e. that levels are separated by "|"
        # and each level should be the sum of its components)
        # is not what we're checking.

        # TODO: split out a function like
        # assert_reported_matches_sum_of_components
        # Would have to be very specific to this setup
        # because of how many different ways it can go wrong
        # e.g. you can have extra components that aren't used,
        # incorrect aggregation

        df_species_aggregate = get_region_sector_sum(
            df_species_internal_consistency_checking_relevant,
            region_level=region_level,
            world_region=world_region,
        ).reorder_levels(df_species.index.names)

        tolerances_species = {}
        for kwarg, value in tolerances[species_total_variable].items():
            if pint is not None and isinstance(value, pint.Quantity):
                if kwarg == "atol":
                    species_units = df_species.index.get_level_values(
                        unit_level
                    ).unique()
                    if len(species_units) > 1:
                        msg = (
                            "Cannot use pint conversion "
                            "if your data contains different units. "
                            f"For {species_total_variable=}, we have {species_units=}"
                        )
                        raise ValueError(msg)

                    tolerances_species[kwarg] = value.to(species_units[0]).m

                elif kwarg == "rtol":
                    tolerances_species[kwarg] = value.to("dimensionless").m

                else:
                    raise NotImplementedError(kwarg)

            else:
                tolerances_species[kwarg] = value

        comparison_species = compare_close(
            left=df_species_total_reported,
            right=df_species_aggregate,
            left_name="reported_total",
            right_name="derived_from_input",
            **tolerances_species,
        )

        if not comparison_species.empty:
            raise InternalConsistencyError(
                differences=comparison_species,
                data_that_was_summed=df_species,
                # Would need something like this to give full details
                # data_that_was_not_summed=data_that_was_not_summed,
                tolerances=tolerances_species,
            )


def to_complete(  # noqa: PLR0913
    indf: pd.DataFrame,
    model_regions: tuple[str, ...],
    unit_level: str = "unit",
    variable_level: str = "variable",
    region_level: str = "region",
    world_region: str = "World",
) -> ToCompleteResult:
    """
    Convert the raw data to complete data

    Parameters
    ----------
    indf
        Data to process

    model_regions
        Model regions to use while reaggregating

    unit_level
        Unit level in the data index

    variable_level
        Variable level in the data index

    region_level
        Region level in the data index

    world_region
        The value used when the data represents the sum over all regions

    Returns
    -------
    :
        To complete result
    """
    assert_only_working_on_variable_unit_region_variations(indf)

    complete_index = get_complete_timeseries_index(
        model_regions=model_regions,
        region_level=region_level,
        variable_level=variable_level,
        world_region=world_region,
    )

    keep = multi_index_lookup(indf, complete_index)
    missing_indexes = get_missing_levels(
        keep.index,  # type: ignore # pandas-stubs confused
        complete_index=complete_index,
        unit_col=unit_level,
    )

    if missing_indexes.empty:
        res = ToCompleteResult(complete=keep, assumed_zero=None)
    else:
        keep_split = split_sectors(keep, middle_level="species")

        species_unit_map = {
            species: unit
            for species, unit in keep_split.index.droplevel(
                keep_split.index.names.difference(["species", unit_level])  # type: ignore #pandas-stubs confused
            )
            .drop_duplicates()
            .reorder_levels(["species", unit_level])
        }

        emissions_mask = missing_indexes.get_level_values(
            variable_level
        ).str.startswith("Emissions")
        missing_indexes_emissions = missing_indexes[emissions_mask]
        missing_indexes_carbon_removal = missing_indexes[~emissions_mask]

        if not missing_indexes_emissions.empty:
            missing_indexes_emissions_split = split_sectors(missing_indexes_emissions)  # type: ignore # type hint is wrong upstream (fix when moving to pandas-openscm)
            zeros_index_split = create_levels_based_on_existing(
                missing_indexes_emissions_split,  # type: ignore # type hint is wrong upstream (fix when moving to pandas-openscm)
                {unit_level: ("species", species_unit_map)},  # type: ignore # type hint is wrong upstream (fix when moving to pandas-openscm)
            )
            zeros_index_emissions: pd.MultiIndex = combine_sectors(  # type: ignore # need to think through type hints for combine_sectors more carefully
                zeros_index_split,  # type: ignore # need to think through type hints for combine_sectors more carefully
                middle_level="species",
            )

        if not missing_indexes_carbon_removal.empty:
            existing_carbon_removal = keep_split[
                keep_split.index.get_level_values("table").str.startswith(
                    "Carbon Removal"
                )
            ]
            if existing_carbon_removal.empty:
                unit = species_unit_map["CO2"]
            else:
                # Use the first, as good as any
                unit = existing_carbon_removal.index.get_level_values("unit")[0]

            zeros_index_carbon_removal = set_levels(
                missing_indexes_carbon_removal, {unit_level: unit}
            )

        if (
            not missing_indexes_emissions.empty
            and not missing_indexes_carbon_removal.empty
        ):
            zeros_index = zeros_index_emissions.append(  # type: ignore # not supported by pandas-stubs
                zeros_index_carbon_removal.reorder_levels(zeros_index_emissions.names)  # type: ignore # not supported by pandas-stubs
            )

        elif not missing_indexes_emissions.empty:
            zeros_index = zeros_index_emissions

        elif not missing_indexes_carbon_removal.empty:
            zeros_index = zeros_index_carbon_removal

        else:
            raise AssertionError

        other_levels_deduped = indf.index.droplevel(
            [variable_level, unit_level, region_level]
        ).drop_duplicates()
        if other_levels_deduped.shape[0] != 1:
            msg = f"Multiple values in other levels:\n{other_levels_deduped=}"
            raise AssertionError(msg)

        extra_levels = {
            level: value
            for level, value in zip(other_levels_deduped.names, other_levels_deduped[0])
        }
        assumed_zero = set_new_single_value_levels(
            pd.DataFrame(
                np.zeros((zeros_index.shape[0], keep.shape[1])),
                columns=keep.columns,
                index=zeros_index,
            ),
            extra_levels,
            copy=False,
        )
        complete = pd.concat([keep, assumed_zero.reorder_levels(keep.index.names)])
        res = ToCompleteResult(complete=complete, assumed_zero=assumed_zero)

    return res


def aggregate_cols(
    df: pd.DataFrame, aggregations: dict[str, list[str]]
) -> pd.DataFrame:
    """
    Aggregate columns

    This is a helper function for [to_gridding_sectors][(m).].

    It does the aggregation in place so we can check that all the columns were used.

    Parameters
    ----------
    df
        Starting [pd.DataFrame][pandas.DataFrame]

    aggregations
        Aggregations to apply

        Each key is the output column,
        each value is the components that contribute to the output column.

    Returns
    -------
    :
        `df` with the aggregations applied
    """
    for aggregate, components in aggregations.items():
        df[aggregate] = df[components].sum(axis="columns")
        df = df.drop(
            # Subtract aggregate in case the aggregate and component have the same name
            list(set(components) - {aggregate}),
            axis="columns",
        )

    return df


def to_gridding_sectors(
    indf: pd.DataFrame, region_level: str = "region", world_region: str = "World"
) -> pd.DataFrame:
    """
    Re-aggregate data to the sectors used for gridding

    Parameters
    ----------
    indf
        Data to re-aggregate

    region_level
        Region level in the data index

    world_region
        The value used when the data represents the sum over all regions

    Returns
    -------
    :
        Data re-aggregated to the gridding sectors
    """
    # Split off the carbon removal tree first
    carbon_removal_mask = indf.index.get_level_values("variable").str.startswith(
        "Carbon Removal"
    )
    carbon_removal = indf[carbon_removal_mask]
    emissions = indf[~carbon_removal_mask]

    # Processing is way easier if we process the DataFrame's a bit first
    emissions_world_mask = (
        emissions.index.get_level_values(region_level) == world_region
    )

    # Data that is at the world level i.e. has no region information
    emissions_sector_df = (
        split_sectors(
            emissions.loc[emissions_world_mask].reset_index("region", drop=True),
            bottom_level="sectors",
        )
        .stack()
        .unstack("sectors")
    )

    # Data with region information
    emissions_region_sector_df = (
        split_sectors(emissions.loc[~emissions_world_mask], bottom_level="sectors")
        .stack()
        .unstack("sectors")
    )

    # CDR information
    emissions_cdr = -1 * update_index_levels_func(
        carbon_removal,
        {"variable": lambda x: x.replace("Carbon Removal", "Emissions|CO2|CDR")},
    )
    emissions_cdr_world_mask = (
        emissions_cdr.index.get_level_values(region_level) == world_region
    )
    if emissions_cdr_world_mask.any():
        raise NotImplementedError

    emissions_region_sector_df_co2_mask = (
        emissions_region_sector_df.index.get_level_values("species") == "CO2"
    )

    emissions_cdr_region_sector_df = convert_unit_like(
        split_sectors(  # type: ignore # pandas-stubs confused
            emissions_cdr.loc[~emissions_cdr_world_mask], bottom_level="sectors"
        )
        .stack()
        .unstack("sectors"),
        emissions_region_sector_df.loc[emissions_region_sector_df_co2_mask],  # type: ignore # pandas-stubs confused
    )

    # Move domestic aviation to the global level,
    # remove it from regional transport
    # and drop the levels we no longer use.
    domestic_aviation_sum = groupby_except(
        emissions_region_sector_df[SECTOR_DOMESTIC_AVIATION],  # type: ignore # issue in pandas-openscm
        region_level,
    ).sum()
    emissions_sector_df["Aircraft"] = (
        emissions_sector_df["Energy|Demand|Bunkers|International Aviation"]
        + domestic_aviation_sum
    )
    emissions_region_sector_df["Energy|Demand|Transportation"] = (
        emissions_region_sector_df["Energy|Demand|Transportation"]
        - emissions_region_sector_df[SECTOR_DOMESTIC_AVIATION]
    )
    emissions_sector_df = emissions_sector_df.drop(
        ["Energy|Demand|Bunkers|International Aviation"], axis="columns"
    )
    emissions_region_sector_df = emissions_region_sector_df.drop(
        [SECTOR_DOMESTIC_AVIATION], axis="columns"
    )

    # Handle the CDR sectors
    carbon_removal_map = {
        "CDR|Enhanced Weathering": "Other Capture and Removal",
        "CDR|Geological Storage|Biomass": "Energy|Supply",
        "CDR|Geological Storage|Direct Air Capture": "Other Capture and Removal",
        "CDR|Ocean": "Other Capture and Removal",
        "CDR|Land Use|Biochar": "AFOLU|Land|Other",
        "CDR|Land Use|Soil Carbon Management": (
            "AFOLU|Land|Land Use and Land-Use Change"
        ),
        # See note above
        # "CDR|Other": "Other Capture and Removal",
    }
    for cdr_sector, emissions_sector in carbon_removal_map.items():
        row_loc = emissions_region_sector_df_co2_mask
        col_loc = emissions_sector
        emissions_region_sector_df.loc[row_loc, col_loc] = (
            emissions_region_sector_df.loc[row_loc, col_loc]
            - emissions_cdr_region_sector_df[cdr_sector]
        )

    # Aggregate
    sector_df_gridding = aggregate_cols(
        emissions_sector_df,  # type: ignore # need to cast first or something
        {
            "International Shipping": ["Energy|Demand|Bunkers|International Shipping"],
            # Not the same as the reporting sector as we have done manipulations above
            "Aircraft": ["Aircraft"],
        },
    )
    region_sector_df_gridding = aggregate_cols(
        emissions_region_sector_df,  # type: ignore # need to cast first or something
        {
            "Agricultural Waste Burning": [
                "AFOLU|Agricultural Waste Burning",
            ],
            "Agriculture": [
                "AFOLU|Agriculture",
                "AFOLU|Land|Harvested Wood Products",
                "AFOLU|Land|Land Use and Land-Use Change",
                "AFOLU|Land|Other",
                "AFOLU|Land|Wetlands",
            ],
            "Energy Sector": ["Energy|Supply"],
            "Forest Burning": ["AFOLU|Land|Fires|Forest Burning"],
            "Grassland Burning": ["AFOLU|Land|Fires|Grassland Burning"],
            "Industrial Sector": [
                "Energy|Demand|Industry",
                "Energy|Demand|Other Sector",
                "Industrial Processes",
                "Other",
            ],
            "Other CDR": [
                "Other Capture and Removal",
            ],
            "Peat Burning": ["AFOLU|Land|Fires|Peat Burning"],
            "Residential Commercial Other": [
                "Energy|Demand|Residential and Commercial and AFOFI"
            ],
            "Solvents Production and Application": ["Product Use"],
            "Transportation Sector": ["Energy|Demand|Transportation"],
            "Waste": ["Waste"],
        },
    )

    emissions_cdr_region_sector_df_gridding = aggregate_cols(
        emissions_cdr_region_sector_df,
        {
            "BECCS": [
                "CDR|Geological Storage|Biomass",
            ],
            "Enhanced Weathering": [
                "CDR|Enhanced Weathering",
            ],
            "Direct Air Capture": [
                "CDR|Geological Storage|Direct Air Capture",
            ],
            "Ocean": [
                "CDR|Ocean",
            ],
            "Biochar": [
                "CDR|Land Use|Biochar",
            ],
            "Soil Carbon Management": [
                "CDR|Land Use|Soil Carbon Management",
            ],
        },
    )

    sector_df_gridding_like_input = combine_sectors(
        set_new_single_value_levels(
            sector_df_gridding.unstack().stack("sectors", future_stack=True),  # type: ignore # pandas-stubs confused
            {region_level: world_region},
        ),
        bottom_level="sectors",
    )
    region_sector_df_gridding_like_input = combine_sectors(
        region_sector_df_gridding.unstack().stack("sectors", future_stack=True),  # type: ignore # pandas-stubs confused
        bottom_level="sectors",
    )
    emissions_cdr_region_sector_df_gridding_like_input = combine_sectors(
        emissions_cdr_region_sector_df_gridding.unstack().stack(
            "sectors",  # type: ignore # pandas-stubs confused
            future_stack=True,
        ),
        bottom_level="sectors",
    )

    res = pd.concat(
        [
            df.reorder_levels(indf.index.names)
            for df in [
                sector_df_gridding_like_input,
                region_sector_df_gridding_like_input,
                emissions_cdr_region_sector_df_gridding_like_input,
            ]
        ]
    )

    return res


@define
class ReaggregatorBasic:
    """
    Reaggregator that follows this module's logic
    """

    model_regions: tuple[str, ...]
    """Model regions to use while reaggregating"""

    region_level: str = "region"
    """Region level in the data index"""

    unit_level: str = "unit"
    """Unit level in the data index"""

    variable_level: str = "variable"
    """Variable level in the data index"""

    world_region: str = "World"
    """
    The value used when the data represents the sum over all regions

    (Having a value for this is odd,
    there should really just be no region level when data is the sum,
    but this is the data format used so we have to follow this convention.)
    """

    internal_consistency_tolerances: (
        Mapping[str, Mapping[str, float]] | Mapping[str, Mapping[str, PINT_SCALAR]]
    ) = field()
    """
    Tolerances to apply when checking the internal consistency of the data
    """

    @internal_consistency_tolerances.default
    def default_tols_internal_consistency(
        self,
    ) -> Mapping[str, Mapping[str, float]] | Mapping[str, Mapping[str, PINT_SCALAR]]:
        """
        Get default tolerances for internal consistency checks
        """
        return get_default_internal_conistency_checking_tolerances()

    def assert_has_all_required_timeseries(self, indf: pd.DataFrame) -> None:
        """
        Assert that the data has all the required timeseries

        Parameters
        ----------
        indf
            Data to check

        Raises
        ------
        NotCompleteError
            `indf` is not complete
        """
        assert_has_all_required_timeseries(
            indf,
            model_regions=self.model_regions,
            world_region=self.world_region,
            region_level=self.region_level,
            variable_level=self.variable_level,
        )

    def assert_is_internally_consistent(self, indf: pd.DataFrame) -> None:
        """
        Assert that the data is internally consistent

        Parameters
        ----------
        indf
            Data to check

        Raises
        ------
        InternalConsistencyError
            The data is not internally consistent
        """
        assert_is_internally_consistent(
            indf,
            model_regions=self.model_regions,
            tolerances=self.internal_consistency_tolerances,
            world_region=self.world_region,
            region_level=self.region_level,
            unit_level=self.unit_level,
            variable_level=self.variable_level,
        )

    def get_internal_consistency_checking_index(self) -> pd.MultiIndex:
        """
        Get the index which selects only data relevant for checking internal consistency

        Returns
        -------
        :
            Internal consistency checking index
        """
        return get_internal_consistency_checking_index(
            model_regions=self.model_regions,
            world_region=self.world_region,
            region_level=self.region_level,
            variable_level=self.variable_level,
        )

    def to_complete(self, raw: pd.DataFrame) -> ToCompleteResult:
        """
        Convert the raw data to complete data

        Parameters
        ----------
        raw
            Raw data

        Returns
        -------
        :
            To complete result
        """
        return to_complete(
            indf=raw,
            model_regions=self.model_regions,
            unit_level=self.unit_level,
            variable_level=self.variable_level,
            region_level=self.region_level,
            world_region=self.world_region,
        )

    def to_gridding_sectors(self, indf: pd.DataFrame) -> pd.DataFrame:
        """
        Re-aggregate data to the sectors used for gridding

        Parameters
        ----------
        indf
            Data to re-aggregate

        Returns
        -------
        :
            Data re-aggregated to the gridding sectors
        """
        return to_gridding_sectors(
            indf=indf, region_level=self.region_level, world_region=self.world_region
        )



"""
Code to support our tests

This is here, rather than in our `tests` directory
because of the issues that come
when you turn your tests into a package using `__init__.py` files
(for details, see https://docs.pytest.org/en/7.1.x/explanation/goodpractices.html#choosing-an-import-mode).
"""

from __future__ import annotations

import functools
import os
import platform
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd
from pandas_openscm.io import load_timeseries_csv

from gcages.exceptions import MissingOptionalDependencyError

if TYPE_CHECKING:
    import pytest

RNG = np.random.default_rng()

AR6_IPS = (
    ("AIM/CGE 2.2", "EN_NPi2020_900f"),
    ("COFFEE 1.1", "EN_NPi2020_400f_lowBECCS"),
    ("GCAM 5.3", "NGFS2_Current Policies"),
    ("IMAGE 3.0", "EN_INDCi2030_3000f"),
    ("MESSAGEix-GLOBIOM 1.0", "LowEnergyDemand_1.3_IPCC"),
    ("MESSAGEix-GLOBIOM_GEI 1.0", "SSP2_openres_lc_50"),
    ("REMIND-MAgPIE 2.1-4.2", "SusDev_SDP-PkBudg1000"),
    ("REMIND-MAgPIE 2.1-4.3", "DeepElec_SSP2_ HighRE_Budg900"),
    ("WITCH 5.0", "CO_Bridge"),
)

KEY_TESTING_MODEL_SCENARIOS = tuple(
    [
        *AR6_IPS,
        # Other special cases
        ("C3IAM 2.0", "2C-hybrid"),
        ("DNE21+ V.14E1", "EMF30_BCOC-EndU"),
    ]
)


def get_key_testing_model_scenario_parameters() -> pytest.MarkDecorator:
    try:
        import pytest
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_key_testing_model_scenario_parameters", requirement="pytest"
        ) from exc

    return pytest.mark.parametrize(
        "model, scenario",
        [(model, scenario) for model, scenario in KEY_TESTING_MODEL_SCENARIOS],
    )


@functools.cache
def get_ar6_all_emissions(
    model: str, scenario: str, processed_ar6_output_data_dir: Path
) -> pd.DataFrame:
    """
    Get all emissions from AR6 for a given model-scenario

    Parameters
    ----------
    model
        Model

    scenario
        Scenario

    processed_ar6_output_data_dir
        Directory in which the AR6 was processed into individual model-scenario files

        (In the repo, see `tests/regression/ar6/convert_ar6_res_to_checking_csvs.py`.)

    Returns
    -------
    :
        All emissions from AR6 for `model`-`scenario`
    """
    filename_emissions = f"ar6_scenarios__{model}__{scenario}__emissions.csv"
    filename_emissions = filename_emissions.replace("/", "_").replace(" ", "_")
    emissions_file = processed_ar6_output_data_dir / filename_emissions

    res = load_timeseries_csv(
        emissions_file,
        index_columns=["model", "scenario", "variable", "region", "unit"],
        out_columns_type=int,
    )

    return res


@functools.cache
def get_ar6_raw_emissions(
    model: str, scenario: str, processed_ar6_output_data_dir: Path
) -> pd.DataFrame:
    """
    Get all raw emissions from AR6 for a given model-scenario

    Parameters
    ----------
    model
        Model

    scenario
        Scenario

    processed_ar6_output_data_dir
        Directory in which the AR6 was processed into individual model-scenario files

        (In the repo, see `tests/regression/ar6/convert_ar6_res_to_checking_csvs.py`.)

    Returns
    -------
    :
        All raw emissions from AR6 for `model`-`scenario`
    """
    try:
        from pandas_indexing.selectors import ismatch
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_ar6_raw_emissions", requirement="pandas_indexing"
        ) from exc

    all_emissions = get_ar6_all_emissions(
        model=model,
        scenario=scenario,
        processed_ar6_output_data_dir=processed_ar6_output_data_dir,
    )
    res: pd.DataFrame = all_emissions.loc[ismatch(variable="Emissions**")].dropna(
        how="all", axis="columns"
    )

    return res


@functools.cache
def get_ar6_harmonised_emissions(
    model: str, scenario: str, processed_ar6_output_data_dir: Path
) -> pd.DataFrame:
    """
    Get all harmonised emissions from AR6 for a given model-scenario

    Parameters
    ----------
    model
        Model

    scenario
        Scenario

    processed_ar6_output_data_dir
        Directory in which the AR6 was processed into individual model-scenario files

        (In the repo, see `tests/regression/ar6/convert_ar6_res_to_checking_csvs.py`.)

    Returns
    -------
    :
        All harmonised emissions from AR6 for `model`-`scenario`
    """
    try:
        from pandas_indexing.selectors import ismatch
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_ar6_harmonised_emissions", requirement="pandas_indexing"
        ) from exc

    all_emissions = get_ar6_all_emissions(
        model=model,
        scenario=scenario,
        processed_ar6_output_data_dir=processed_ar6_output_data_dir,
    )
    res: pd.DataFrame = all_emissions.loc[ismatch(variable="**Harmonized**")].dropna(
        how="all", axis="columns"
    )

    return res


@functools.cache
def get_ar6_infilled_emissions(
    model: str, scenario: str, processed_ar6_output_data_dir: Path
) -> pd.DataFrame:
    """
    Get all infilled emissions from AR6 for a given model-scenario

    Parameters
    ----------
    model
        Model

    scenario
        Scenario

    processed_ar6_output_data_dir
        Directory in which the AR6 output was processed into model-scenario files

        (In the repo, see `tests/regression/ar6/convert_ar6_res_to_checking_csvs.py`.)

    Returns
    -------
    :
        All infilled emissions from AR6 for `model`-`scenario`
    """
    try:
        from pandas_indexing.selectors import ismatch
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_ar6_infilled_emissions", requirement="pandas_indexing"
        ) from exc

    all_emissions = get_ar6_all_emissions(
        model=model,
        scenario=scenario,
        processed_ar6_output_data_dir=processed_ar6_output_data_dir,
    )
    res: pd.DataFrame = all_emissions.loc[ismatch(variable="**Infilled**")].dropna(
        how="all", axis="columns"
    )

    return res


@functools.cache
def get_ar6_temperature_outputs(
    model: str, scenario: str, processed_ar6_output_data_dir: Path, dropna: bool = True
) -> pd.DataFrame:
    """
    Get temperature outputs we've downloaded from AR6 for a given model-scenario

    Parameters
    ----------
    model
        Model

    scenario
        Scenario

    processed_ar6_output_data_dir
        Directory in which the AR6 output was processed into model-scenario files

        (In the repo, see `tests/regression/ar6/convert_ar6_res_to_checking_csvs.py`.)

    dropna
        Drop time columns that only contain NaN

    Returns
    -------
    :
        All temperature outputs we've downloaded from AR6 for `model`-`scenario`
    """
    filename_temperatures = f"ar6_scenarios__{model}__{scenario}__temperatures.csv"
    filename_temperatures = filename_temperatures.replace("/", "_").replace(" ", "_")
    temperatures_file = processed_ar6_output_data_dir / filename_temperatures

    res = load_timeseries_csv(
        temperatures_file,
        index_columns=["model", "scenario", "variable", "region", "unit"],
        out_columns_type=int,
    )
    if dropna:
        res = res.dropna(axis="columns", how="all")

    return res


@functools.cache
def get_ar6_metadata_outputs(
    model: str,
    scenario: str,
    ar6_output_data_dir: Path,
    filename: str = "AR6_Scenarios_Database_metadata_indicators_v1.1_meta.csv",
) -> pd.DataFrame:
    """
    Get metadata from AR6 for a given model-scenario

    Parameters
    ----------
    model
        Model

    scenario
        Scenario

    ar6_output_data_dir
        Directory in which the AR6 output was saved

    Returns
    -------
    :
        Metadata from AR6 for `model`-`scenario`
    """
    res = load_timeseries_csv(
        ar6_output_data_dir / filename,
        lower_column_names=False,
        index_columns=["Model", "Scenario"],
    ).loc[[(model, scenario)]]

    res.index = res.index.rename({"Model": "model", "Scenario": "scenario"})

    return res


def guess_magicc_exe_path() -> Path:
    """
    Guess the path to the MAGICC executable

    Uses the `MAGICC_EXECUTABLE_7` environment variable.
    If that isn't set, it guesses.

    Returns
    -------
    :
        Path to the MAGICC executable

    Raises
    ------
    FileNotFoundError
        The guessed path to the MAGICC executable does not exist
    """
    env_var = os.environ.get("MAGICC_EXECUTABLE_7", None)
    if env_var is not None:
        return Path(env_var)

    guess = None
    guess_path = (
        Path(__file__).parents[2]
        / "tests"
        / "regression"
        / "ar6"
        / "ar6-workflow-inputs"
        / "magicc-v7.5.3"
        / "bin"
    )
    if platform.system() == "Darwin":
        if platform.processor() == "arm":
            guess = guess_path / "magicc-darwin-arm64"

    elif platform.system() == "Linux":
        guess = guess_path / "magicc"

    elif platform.system() == "Windows":
        guess = guess_path / "magicc.exe"

    if guess is not None:
        if guess.exists():
            return guess

        msg = f"Guessed that the MAGICC executable was in: {guess}"
        raise FileNotFoundError(msg)

    msg = "No guess about where the MAGICC executable is for your system"
    raise FileNotFoundError(msg)


def assert_frame_equal(
    res: pd.DataFrame, exp: pd.DataFrame, rtol: float = 1e-8, **kwargs: Any
) -> None:
    """
    Assert two [pd.DataFrame][pandas.DataFrame]'s are equal.

    This is a very thin wrapper around
    [pd.testing.assert_frame_equal][pandas.testing.assert_frame_equal]
    that makes some use of [pandas_indexing][]
    to give slightly nicer and clearer errors.

    Parameters
    ----------
    res
        Result

    exp
        Expected value

    rtol
        Relative tolerance

    **kwargs
        Passed to [pd.testing.assert_frame_equal][pandas.testing.assert_frame_equal]

    Raises
    ------
    AssertionError
        The frames aren't equal
    """
    try:
        from pandas_indexing.core import uniquelevel
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "assert_frame_equal", requirement="pandas_indexing"
        ) from exc

    for idx_name in res.index.names:
        idx_diffs = uniquelevel(res, idx_name).symmetric_difference(
            uniquelevel(exp, idx_name)
        )
        if not idx_diffs.empty:
            msg = f"Differences in the {idx_name} (res on the left): {idx_diffs=}"
            raise AssertionError(msg)

    pd.testing.assert_frame_equal(
        res.reorder_levels(exp.index.names).T,
        exp.T,
        check_like=True,
        check_exact=False,
        rtol=rtol,
        **kwargs,
    )


# TODO: move into pandas_openscm
def compare_close(
    left: pd.DataFrame,
    right: pd.DataFrame,
    left_name: str,
    right_name: str,
    rtol: float = 1e-8,
    **kwargs: Any,
) -> pd.DataFrame:
    left_stacked = left.stack()
    left_stacked.name = left_name

    right_stacked = right.stack()
    right_stacked.name = right_name

    left_stacked_aligned, right_stacked_aligned = left_stacked.align(right_stacked)
    differences_locator = ~np.isclose(
        left_stacked_aligned, right_stacked_aligned, rtol=rtol, **kwargs
    )

    res = pd.concat(
        [
            left_stacked_aligned[differences_locator],
            right_stacked_aligned[differences_locator],
        ],
        axis="columns",
    )

    return res


def get_variable_unit_default(v: str) -> str:
    if v.startswith("Carbon Removal"):
        return "Mt CO2/yr"

    species = v.split("|")[1]
    unit_map = {
        "BC": "Mt BC/yr",
        "CH4": "Mt CH4/yr",
        "CO": "Mt CO/yr",
        "CO2": "Gt C/yr",
        "N2O": "kt N2O/yr",
        "NH3": "Mt NH3/yr",
        "NOx": "Mt NO2/yr",
        "OC": "Mt OC/yr",
        "Sulfur": "Mt SO2/yr",
        "VOC": "Mt VOC/yr",
    }
    return unit_map[species]



"""
Manipulation of the index of [pd.DataFrame][pandas.DataFrame]'s
"""

# TOOD: put all of this in pandas-openscm
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, TypeVar

import pandas as pd

from gcages.exceptions import MissingOptionalDependencyError

if TYPE_CHECKING:
    P = TypeVar("P", pd.DataFrame, pd.Series[Any])


def set_new_single_value_levels(  # noqa: D103
    pandas_obj: P,
    levels_to_set: dict[str, Any],  # indicate not a collection somehow
    copy: bool = True,
) -> P:
    # TODO: move to pandas-openscm
    # TODO: split out method that just works on MultiIndex
    if copy:
        pandas_obj = pandas_obj.copy()

    new_names = levels_to_set.keys()
    new_values = levels_to_set.values()

    if not isinstance(pandas_obj.index, pd.MultiIndex):
        raise TypeError(pandas_obj.index)

    pandas_obj.index = pd.MultiIndex(
        codes=[
            *pandas_obj.index.codes,  # type: ignore #  not sure why check above isn't working
            *([[0] * pandas_obj.index.shape[0]] * len(new_values)),  # type: ignore # fix when moving to pandas-openscm
        ],
        levels=[*pandas_obj.index.levels, *[pd.Index([value]) for value in new_values]],  # type: ignore # fix when moving to pandas-openscm
        names=[*pandas_obj.index.names, *new_names],  # type: ignore # fix when moving to pandas-openscm
    )

    return pandas_obj


def create_levels_based_on_existing(  # noqa: D103
    ini: pd.MultiIndex,
    create_from: dict[Any, tuple[str, Callable[[Any], Any]]],
    remove_unused_levels: bool = True,
) -> pd.MultiIndex:
    # TODO: move to pandas-openscm
    if remove_unused_levels:
        ini = ini.remove_unused_levels()  # type: ignore

    levels: list[pd.Index[Any]] = list(ini.levels)
    codes: list[list[int]] = list(ini.codes)
    names: list[str] = list(ini.names)

    for level, (source, updater) in create_from.items():
        if source not in ini.names:
            msg = (
                f"{source} is not available in the index. Available levels: {ini.names}"
            )
            raise KeyError(msg)

        source_idx = ini.names.index(source)
        new_level = ini.levels[source_idx].map(updater)
        if not new_level.has_duplicates:
            # Fast route: no clashes so no need to update the codes
            # or do anything
            new_codes = ini.codes[source_idx]

        else:
            # Slow route: have to update the codes too
            dup_level = ini.get_level_values(source).map(updater)
            new_level = new_level.unique()
            new_codes = new_level.get_indexer(dup_level)  # type: ignore

        levels.append(new_level)
        codes.append(new_codes)
        names.append(level)

    res = pd.MultiIndex(levels=levels, codes=codes, names=names)

    return res


def split_sectors(  # noqa: PLR0913
    indf: pd.DataFrame,
    dropna: bool = True,
    level_to_split: str = "variable",
    top_level: str = "table",
    middle_level: str = "species",
    bottom_level: str = "sectors",
    level_separator: str = "|",
) -> pd.DataFrame:
    """
    Split the input [pd.DataFrame][pandas.DataFrame]'s level to split into sectors

    Any levels beyond three are all left in `bottom_level`
    in the output.

    This is the inverse of [combine_sectors][(m).].

    Parameters
    ----------
    indf
        Input data

    dropna
        Should levels which have NaNs after splitting be dropped?

    level_to_split
        The level to split

    top_level
        Name of the top level after the split

    middle_level
        Name of the middle level after the split

    bottom_level
        Name of the bottom level after the split

    level_separator
        Separator between levels in `level_to_split`

    Returns
    -------
    :
        `indf` with `level_to_split` split into three levels
        with the given names

    Examples
    --------
    >>> indf = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions|CO2|sector", "t / yr"),
    ...             ("sa", "Emissions|CO2|sector|sub sector", "t / yr"),
    ...             ("sb", "Emissions|CH4|sector|sub|sub-sub|sub-sub-sub", "kg / yr"),
    ...             ("sb", "Emissions|CO2|sector", "t / yr"),
    ...         ],
    ...         names=["scenario", "variable", "unit"],
    ...     ),
    ... )
    >>> split_sectors(indf)  # doctest: +NORMALIZE_WHITESPACE
                                                                       2015  2100
    scenario unit    table     species sectors
    sa       t / yr  Emissions CO2     sector                           1.0   2.0
                                       sector|sub sector                3.0   2.0
    sb       kg / yr Emissions CH4     sector|sub|sub-sub|sub-sub-sub   1.3   2.2
             t / yr  Emissions CO2     sector                           3.4   2.1

    >>> indf_funky = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions-CO2-sector", "t / yr"),
    ...             ("sa", "Emissions-CO2-sector-sub sector", "t / yr"),
    ...             ("sb", "Emissions-CH4-sector-sub-sector-transport", "kg / yr"),
    ...             ("sb", "Emissions-CO2-sector", "t / yr"),
    ...         ],
    ...         names=["scenario", "vv", "unit"],
    ...     ),
    ... )
    >>> split_sectors(
    ...     indf_funky,
    ...     level_to_split="vv",
    ...     top_level="t",
    ...     middle_level="m",
    ...     bottom_level="b",
    ...     level_separator="-",
    ... )  # doctest: +NORMALIZE_WHITESPACE
                                                                2015  2100
    scenario unit    t         m   b
    sa       t / yr  Emissions CO2 sector                        1.0   2.0
                                   sector-sub sector             3.0   2.0
    sb       kg / yr Emissions CH4 sector-sub-sector-transport   1.3   2.2
             t / yr  Emissions CO2 sector                        3.4   2.1
    """
    try:
        from pandas_indexing.core import extractlevel
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "split_sectors", requirement="pandas_indexing"
        ) from exc

    kwargs = {
        level_to_split: level_separator.join(
            [
                "{" + f"{level}" + "}"
                for level in [top_level, middle_level, bottom_level]
            ]
        )
    }

    return extractlevel(indf, dropna=dropna, **kwargs)  # type: ignore


def split_species(  # noqa: PLR0913
    indf: pd.DataFrame,
    dropna: bool = True,
    level_to_split: str = "variable",
    top_level: str = "table",
    bottom_level: str = "species",
    level_separator: str = "|",
) -> pd.DataFrame:
    """
    Split the input [pd.DataFrame][pandas.DataFrame]'s level to split into species

    Any levels beyond two are all left in `bottom_level` in the output.

    This is the inverse of [combine_species][(m).].

    Parameters
    ----------
    indf
        Input data

    dropna
        Should levels which have NaNs after splitting be dropped?

    level_to_split
        The level to split

    top_level
        Name of the top level after the split

    bottom_level
        Name of the bottom level after the split

    level_separator
        Separator between levels in `level_to_split`

    Returns
    -------
    :
        `indf` with `level_to_split` split into two levels
        with the given names

    Examples
    --------
    >>> indf = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions|CO2", "t / yr"),
    ...             ("sa", "Emissions|CH4", "t / yr"),
    ...             ("sb", "Emissions|CH4|sector|sub", "kg / yr"),
    ...             ("sb", "Emissions|N2O", "t / yr"),
    ...         ],
    ...         names=["scenario", "variable", "unit"],
    ...     ),
    ... )
    >>> split_species(indf)  # doctest: +NORMALIZE_WHITESPACE
                                               2015  2100
    scenario unit    table     species
    sa       t / yr  Emissions CO2              1.0   2.0
                               CH4              3.0   2.0
    sb       kg / yr Emissions CH4|sector|sub   1.3   2.2
             t / yr  Emissions N2O              3.4   2.1
    >>> indf_funky = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions-CO2", "t / yr"),
    ...             ("sa", "Emissions-CO2-sector-sub sector", "t / yr"),
    ...             ("sb", "Emissions-CH4", "kg / yr"),
    ...             ("sb", "Emissions-N2O", "t / yr"),
    ...         ],
    ...         names=["scenario", "vv", "unit"],
    ...     ),
    ... )
    >>> split_species(
    ...     indf_funky,
    ...     level_to_split="vv",
    ...     top_level="t",
    ...     bottom_level="b",
    ...     level_separator="-",
    ... )  # doctest: +NORMALIZE_WHITESPACE
                                                      2015  2100
    scenario unit    t         b
    sa       t / yr  Emissions CO2                     1.0   2.0
                               CO2-sector-sub sector   3.0   2.0
    sb       kg / yr Emissions CH4                     1.3   2.2
             t / yr  Emissions N2O                     3.4   2.1
    """
    try:
        from pandas_indexing.core import extractlevel
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "split_species", requirement="pandas_indexing"
        ) from exc

    kwargs = {
        level_to_split: level_separator.join(
            ["{" + f"{level}" + "}" for level in [top_level, bottom_level]]
        )
    }

    return extractlevel(indf, dropna=dropna, **kwargs)  # type: ignore


def combine_sectors(  # noqa: PLR0913
    indf: pd.DataFrame,  # Could open this out as can also work on Series and MultiIndex
    drop: bool = True,
    combined_level: str = "variable",
    top_level: str = "table",
    middle_level: str = "species",
    bottom_level: str = "sectors",
    level_separator: str = "|",
) -> pd.DataFrame:
    """
    Combine the input [pd.DataFrame][pandas.DataFrame]'s levels

    This assumes that you want to combine three levels.

    This is the inverse of [split_sectors][(m).].

    Parameters
    ----------
    indf
        Input data

    drop
        Should the combined levels be dropped?

    combined_level
        The name of the combined level

    top_level
        Name of the top level in the combined output

    middle_level
        Name of the middle level in the combined output

    bottom_level
        Name of the bottom level in the combined output

    level_separator
        Separator between levels in `combined_level`

    Returns
    -------
    :
        `indf` with the given levels combined into `combined_level`

    Examples
    --------
    >>> indf = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions", "CO2", "sector", "t / yr"),
    ...             ("sa", "Emissions", "CO2", "sector|sub sector", "t / yr"),
    ...             (
    ...                 "sb",
    ...                 "Emissions",
    ...                 "CH4",
    ...                 "sector|sub|sub-sub|sub-sub-sub",
    ...                 "kg / yr",
    ...             ),
    ...             ("sb", "Emissions", "CO2", "sector", "t / yr"),
    ...         ],
    ...         names=["scenario", "table", "species", "sectors", "unit"],
    ...     ),
    ... )
    >>> combine_sectors(indf)  # doctest: +NORMALIZE_WHITESPACE
                                                                   2015  2100
    scenario unit    variable
    sa       t / yr  Emissions|CO2|sector                           1.0   2.0
                     Emissions|CO2|sector|sub sector                3.0   2.0
    sb       kg / yr Emissions|CH4|sector|sub|sub-sub|sub-sub-sub   1.3   2.2
             t / yr  Emissions|CO2|sector                           3.4   2.1
    >>> indf_funky = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions", "CO2", "sector", "t / yr"),
    ...             ("sa", "Emissions", "CO2", "sector-sub sector", "t / yr"),
    ...             ("sb", "Emissions", "CH4", "sector-sub-sub", "kg / yr"),
    ...             ("sb", "Emissions", "CO2", "sector", "t / yr"),
    ...         ],
    ...         names=["scenario", "prefix", "gas", "sector", "unit"],
    ...     ),
    ... )
    >>> combine_sectors(
    ...     indf_funky,
    ...     combined_level="vv",
    ...     top_level="prefix",
    ...     middle_level="gas",
    ...     bottom_level="sector",
    ...     level_separator="-",
    ... )  # doctest: +NORMALIZE_WHITESPACE
                                                      2015  2100
    scenario unit    vv
    sa       t / yr  Emissions-CO2-sector              1.0   2.0
                     Emissions-CO2-sector-sub sector   3.0   2.0
    sb       kg / yr Emissions-CH4-sector-sub-sub      1.3   2.2
             t / yr  Emissions-CO2-sector              3.4   2.1
    """
    try:
        from pandas_indexing.core import formatlevel
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "combine_sectors", requirement="pandas_indexing"
        ) from exc

    kwargs = {
        combined_level: level_separator.join(
            [
                "{" + f"{level}" + "}"
                for level in [top_level, middle_level, bottom_level]
            ]
        )
    }

    return formatlevel(indf, drop=drop, **kwargs)  # type: ignore


def combine_species(  # noqa: PLR0913
    indf: pd.DataFrame,
    drop: bool = True,
    combined_level: str = "variable",
    top_level: str = "table",
    bottom_level: str = "species",
    level_separator: str = "|",
) -> pd.DataFrame:
    """
    Combine the input [pd.DataFrame][pandas.DataFrame]'s levels

    This assumes that you want to combine two levels.

    This is the inverse of [split_species][(m).].

    Parameters
    ----------
    indf
        Input data

    drop
        Should the combined levels be dropped?

    combined_level
        The name of the combined level

    top_level
        Name of the top level in the combined output

    bottom_level
        Name of the bottom level in the combined output

    level_separator
        Separator between levels in `combined_level`

    Returns
    -------
    :
        `indf` with the given levels combined into `combined_level`

    Examples
    --------
    >>> indf = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions", "CO2", "t / yr"),
    ...             ("sa", "Emissions", "CO2", "t / yr"),
    ...             ("sb", "Emissions", "CH4", "kg / yr"),
    ...             ("sb", "Emissions", "CO2", "t / yr"),
    ...         ],
    ...         names=["scenario", "table", "species", "unit"],
    ...     ),
    ... )
    >>> combine_species(indf)  # doctest: +NORMALIZE_WHITESPACE
                                    2015  2100
    scenario unit    variable
    sa       t / yr  Emissions|CO2   1.0   2.0
                     Emissions|CO2   3.0   2.0
    sb       kg / yr Emissions|CH4   1.3   2.2
             t / yr  Emissions|CO2   3.4   2.1
    >>> indf_funky = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.3, 2.2],
    ...         [3.4, 2.1],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "Emissions", "CO2", "t / yr"),
    ...             ("sa", "Emissions", "CO2-sector-sub sector", "t / yr"),
    ...             ("sb", "Emissions", "CH4-sector-sub-sub", "kg / yr"),
    ...             ("sb", "Emissions", "CO2", "t / yr"),
    ...         ],
    ...         names=["scenario", "prefix", "gas", "unit"],
    ...     ),
    ... )
    >>> combine_species(
    ...     indf_funky,
    ...     combined_level="vv",
    ...     top_level="prefix",
    ...     bottom_level="gas",
    ...     level_separator="-",
    ... )  # doctest: +NORMALIZE_WHITESPACE
                                                      2015  2100
    scenario unit    vv
    sa       t / yr  Emissions-CO2                     1.0   2.0
                     Emissions-CO2-sector-sub sector   3.0   2.0
    sb       kg / yr Emissions-CH4-sector-sub-sub      1.3   2.2
             t / yr  Emissions-CO2                     3.4   2.1
    """
    try:
        from pandas_indexing.core import formatlevel
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "combine_species", requirement="pandas_indexing"
        ) from exc

    kwargs = {
        combined_level: level_separator.join(
            ["{" + f"{level}" + "}" for level in [top_level, bottom_level]]
        )
    }

    return formatlevel(indf, drop=drop, **kwargs)  # type: ignore



"""
Pre-processing part of the workflow

This is extremely fiddly
because of the way the data is reported,
which is frankly, a mess
because of how it blends data
that is a regional-sum with data that has regional detail
and how the variable name is a blend of different bits of information
(species, sectoral information etc.)
with no easy way to decode what is what using a machine
(you have to hardcode lots of edge cases
e.g. Emissions|CO2|Energy is "Emissions", then the species then the sector
but Emissions|HFC|HFC245 is "Emissions" then the "HFC" string then the species,
i.e. completely different information is provided after each "|").

This module implements the logic for this processing.
The complexity comes in the re-aggregation
([gcages.cmip7_scenariomip.pre_processing.reaggregation][]),
which has to handle converting from whatever is reported
(and a huge amount of different possibilities have to be supported)
to the sectors used for gridding.
From there, the workflow can be standardised
(as is done in
[pre_processor.do_pre_processing][(m).]).
"""

from __future__ import annotations

from gcages.cmip7_scenariomip.pre_processing.pre_processor import (
    CMIP7ScenarioMIPPreProcessingResult,
    CMIP7ScenarioMIPPreProcessor,
    ReaggregatorLike,
)
from gcages.cmip7_scenariomip.pre_processing.reaggregation import (
    ReaggregatorBasic,
    ToCompleteResult,
)

__all__ = [
    "CMIP7ScenarioMIPPreProcessingResult",
    "CMIP7ScenarioMIPPreProcessor",
    "ReaggregatorBasic",
    "ReaggregatorLike",
    "ToCompleteResult",
]



"""
Post-processing part of the AR6 workflow
"""

from __future__ import annotations

import multiprocessing
from typing import TypeVar

import numpy as np
import pandas as pd
from attrs import define
from pandas_openscm.grouping import (
    fix_index_name_after_groupby_quantile,
    groupby_except,
)
from pandas_openscm.index_manipulation import update_index_levels_func

from gcages.assertions import (
    assert_data_is_all_numeric,
    assert_has_data_for_times,
    assert_has_index_levels,
    assert_index_is_multiindex,
)
from gcages.index_manipulation import set_new_single_value_levels
from gcages.post_processing import PostProcessingResult

T = TypeVar("T")


def get_temperatures_in_line_with_assessment(
    raw_temperatures: pd.DataFrame,
    assessment_median: float,
    assessment_time_period: tuple[int, ...],
    assessment_pre_industrial_period: tuple[int, ...],
    group_cols: list[str],
) -> pd.DataFrame:
    """
    Get temperatures in line with the historical assessment

    Parameters
    ----------
    raw_temperatures
        Raw temperatures

    assessment_median
        Median of the assessment to match

    assessment_time_period
        Time period over which the assessment applies

    assessment_pre_industrial_period
        Pre-industrial period used for the assessment

    group_cols
        Columns to use when grouping `raw_temperatures`

    Returns
    -------
    :
        Temperatures,
        adjusted so their medians are in line with the historical assessment.
    """
    # TODO: move to pandas-openscm
    pre_industrial_mean = raw_temperatures.loc[
        :, list(assessment_pre_industrial_period)
    ].mean(axis="columns")
    rel_pi_temperatures = raw_temperatures.subtract(pre_industrial_mean, axis="rows")  # type: ignore # pandas-stubs confused

    assessment_period_median = (
        rel_pi_temperatures.loc[:, list(assessment_time_period)]
        .mean(axis="columns")
        .groupby(group_cols)
        .median()
    )
    res = (
        rel_pi_temperatures.subtract(assessment_period_median, axis="rows")  # type: ignore # pandas-stubs confused
        + assessment_median
    )
    # Checker:
    # res.loc[:, list(assessment_time_period)].mean(axis="columns").groupby( ["model", "scenario"]).median()  # noqa: E501

    return res


def get_exceedance_probabilities_over_time(  # noqa: D103
    temperatures: pd.DataFrame,
    exceedance_thresholds_of_interest: tuple[float, ...],
    group_cols: list[str],
    unit_col: str,
    groupby_except_levels: list[str] | str,
) -> pd.DataFrame:
    # TODO: Move to pandas-openscm
    n_runs_per_group = temperatures.groupby(group_cols).count()

    # TODO: add get_index_value to pandas-openscm with keyword `expect_singular`
    temperature_unit_l = temperatures.index.get_level_values(unit_col).unique().tolist()
    if len(temperature_unit_l) > 1:
        raise AssertionError(temperature_unit_l)
    temperature_unit = temperature_unit_l[0]

    exceedance_probs_l = []
    for thresh in exceedance_thresholds_of_interest:
        exceedance_prob_transient = update_index_levels_func(
            groupby_except(temperatures > thresh, groupby_except_levels)
            .sum()
            .divide(n_runs_per_group)
            * 100,
            dict(
                variable=lambda x: "Exceedance probability",
                unit=lambda x: "%",
            ),
        )
        exceedance_prob_transient = set_new_single_value_levels(
            exceedance_prob_transient,
            {"threshold": thresh, "threshold_unit": temperature_unit},
            copy=False,
        )

        exceedance_probs_l.append(exceedance_prob_transient)

    exceedance_probs = pd.concat(exceedance_probs_l)

    return exceedance_probs


def get_exceedance_probabilities(  # noqa: D103
    temperatures: pd.DataFrame,
    exceedance_thresholds_of_interest: tuple[float, ...],
    group_cols: list[str],
    unit_col: str,
    groupby_except_levels: list[str] | str,
) -> pd.Series[float]:
    # TODO: Move to pandas-openscm
    # Have to do it this way because we don't want the columns
    n_runs_per_group = temperatures.index.droplevel(
        temperatures.index.names.difference(group_cols)  # type: ignore # pandas-stubs confused
    ).value_counts()

    # TODO: add get_index_value to pandas-openscm with keyword `expect_singular`
    temperature_unit_l = temperatures.index.get_level_values(unit_col).unique().tolist()
    if len(temperature_unit_l) > 1:
        raise AssertionError(temperature_unit_l)
    temperature_unit = temperature_unit_l[0]

    peak_warming = temperatures.max(axis="columns")

    exceedance_probs_l = []
    for thresh in exceedance_thresholds_of_interest:
        exceedance_prob = update_index_levels_func(
            groupby_except(peak_warming > thresh, groupby_except_levels)  # type: ignore # error in pandas-openscm
            .sum()
            .divide(n_runs_per_group)
            * 100,
            dict(
                variable=lambda x: "Exceedance probability",
                unit=lambda x: "%",
            ),
        )
        exceedance_prob = set_new_single_value_levels(
            exceedance_prob,
            {"threshold": thresh, "threshold_unit": temperature_unit},
            copy=False,
        )

        exceedance_probs_l.append(exceedance_prob)

    exceedance_probs: pd.Series[float] = pd.concat(exceedance_probs_l)  # type: ignore # pandas-stubs out of date

    return exceedance_probs


def categorise_scenarios(
    peak_warming_quantiles: pd.DataFrame,
    eoc_warming_quantiles: pd.DataFrame,
    group_levels: list[str],
    quantile_level: str,
) -> pd.Series[str]:
    """
    Categorise scenarios

    Parameters
    ----------
    peak_warming_quantiles
        Peak warming quantiles

    eoc_warming_quantiles
        End of century warming quantiles

    group_levels
        Levels of the input indexes to group the results by

        In other words, each unique combination of values in `group_levels`
        will get its own category.

        Typically, this is something like `["model", "scenario", "climate_model"]`

    quantile_level
        The level in `peak_warming_quantiles` and `eoc_warming_quantiles`
        that holds information about the quantile of each value.

    Returns
    -------
    :
        Scenario categorisation
    """
    index = peak_warming_quantiles.index.droplevel(
        peak_warming_quantiles.index.names.difference(group_levels)  # type: ignore # pandas-stubs confused
    ).unique()

    peak_warming_quantiles_use = peak_warming_quantiles.reset_index(
        peak_warming_quantiles.index.names.difference([*group_levels, quantile_level]),  # type: ignore # pandas-stubs confused
        drop=True,
    ).unstack(quantile_level)
    eoc_warming_quantiles_use = eoc_warming_quantiles.reset_index(
        eoc_warming_quantiles.index.names.difference([*group_levels, quantile_level]),  # type: ignore # pandas-stubs confused
        drop=True,
    ).unstack(quantile_level)

    category_names = pd.Series("C8: exceed warming of 4°C (>=50%)", index=index)
    category_names[peak_warming_quantiles_use[0.5] < 4.0] = (  # noqa: PLR2004
        "C7: limit warming to 4°C (>50%)"
    )
    category_names[peak_warming_quantiles_use[0.5] < 3.0] = (  # noqa: PLR2004
        "C6: limit warming to 3°C (>50%)"
    )
    category_names[peak_warming_quantiles_use[0.5] < 2.5] = (  # noqa: PLR2004
        "C5: limit warming to 2.5°C (>50%)"
    )
    category_names[peak_warming_quantiles_use[0.5] < 2.0] = (  # noqa: PLR2004
        "C4: limit warming to 2°C (>50%)"
    )
    category_names[peak_warming_quantiles_use[0.67] < 2.0] = (  # noqa: PLR2004
        "C3: limit warming to 2°C (>67%)"
    )
    category_names[
        (peak_warming_quantiles_use[0.33] > 1.5)  # noqa: PLR2004
        & (eoc_warming_quantiles_use[0.5] < 1.5)  # noqa: PLR2004
    ] = "C2: return warming to 1.5°C (>50%) after a high overshoot"
    category_names[
        (peak_warming_quantiles_use[0.33] <= 1.5)  # noqa: PLR2004
        & (eoc_warming_quantiles_use[0.5] < 1.5)  # noqa: PLR2004
    ] = "C1: limit warming to 1.5°C (>50%) with no or limited overshoot"

    category_names = set_new_single_value_levels(
        category_names, {"metric": "category_name"}, copy=False
    )
    categories = update_index_levels_func(
        category_names.apply(lambda x: x.split(":")[0]),
        {"metric": lambda x: "category"},
    )
    out: pd.Series[str] = pd.concat([category_names, categories])  # type: ignore # pandas-stubs out of date

    return out


@define
class AR6PostProcessor:
    """
    Post-processor that follows the same logic as was used in AR6

    If you want exactly the same behaviour as in AR6,
    initialise using [`from_ar6_config`][(c)]
    """

    gsat_assessment_median: float
    """
    Median of the GSAT assessment
    """

    gsat_assessment_time_period: tuple[int, ...]
    """
    Time period over which the GSAT assessment applies
    """

    gsat_assessment_pre_industrial_period: tuple[int, ...]
    """
    Pre-industrial time period used for the GSAT assessment
    """

    quantiles_of_interest: tuple[float, ...]
    """
    Quantiles to include in output
    """

    exceedance_thresholds_of_interest: tuple[float, ...]
    """
    Thresholds of interest for calculating exceedance probabilities
    """

    raw_gsat_variable_in: str
    """
    Name of the variable that contains raw temperature output in the input

    The temperature output should be global-mean surface air temperature (GSAT).
    """

    assessed_gsat_variable: str
    """
    Name of the output variable that will contain temperature output

    This temperature output is in line with the (AR6) assessed historical warming.
    """

    run_checks: bool = True
    """
    If `True`, run checks on both input and output data

    If you are sure about your workflow,
    you can disable the checks to speed things up
    (but we don't recommend this unless you really
    are confident about what you're doing).
    """

    progress: bool = True
    """
    Should progress bars be shown for each operation where they make sense?
    """

    n_processes: int | None = multiprocessing.cpu_count()
    """
    Number of processes to use for parallel processing.

    Set to `None` to process in serial.
    """

    def __call__(self, in_df: pd.DataFrame) -> PostProcessingResult:
        """
        Do the post-processing

        Parameters
        ----------
        in_df
            Data to post-process

        Returns
        -------
        timeseries, metadata :
            Post-processed results

            These are both timeseries as well as scenario-level metadata.
        """
        if self.run_checks:
            assert_index_is_multiindex(in_df)
            assert_has_index_levels(
                in_df, ["variable", "unit", "model", "scenario", "climate_model"]
            )
            assert_data_is_all_numeric(in_df)
            assert_has_data_for_times(
                in_df, name="in_df", times=[2100], allow_nan=False
            )

            if self.raw_gsat_variable_in not in in_df.index.get_level_values(
                "variable"
            ):
                msg = (
                    f"{self.raw_gsat_variable_in} must be provided. "
                    f"Received: {in_df.index.get_level_values('variable')=}"
                )
                raise AssertionError(msg)

        temperatures_in_line_with_assessment = update_index_levels_func(
            get_temperatures_in_line_with_assessment(
                in_df.loc[
                    in_df.index.get_level_values("variable").isin(
                        [self.raw_gsat_variable_in]
                    )
                ],
                assessment_median=self.gsat_assessment_median,
                assessment_time_period=self.gsat_assessment_time_period,
                assessment_pre_industrial_period=self.gsat_assessment_pre_industrial_period,
                group_cols=["climate_model", "model", "scenario"],
            ),
            {"variable": lambda x: self.assessed_gsat_variable},
        )
        temperatures_in_line_with_assessment_quantiles = (
            fix_index_name_after_groupby_quantile(
                groupby_except(
                    temperatures_in_line_with_assessment,
                    "run_id",
                ).quantile(self.quantiles_of_interest),  # type: ignore # pandas-stubs confused
                new_name="quantile",
            )
        )
        exceedance_probabilities_over_time = get_exceedance_probabilities_over_time(
            temperatures_in_line_with_assessment,
            exceedance_thresholds_of_interest=self.exceedance_thresholds_of_interest,
            group_cols=["model", "scenario", "climate_model"],
            unit_col="unit",
            groupby_except_levels="run_id",
        )

        # TODO: move pandas-openscm.max to pandas-openscm
        peak_warming = set_new_single_value_levels(
            temperatures_in_line_with_assessment.max(axis="columns"), {"metric": "max"}
        )
        peak_warming_quantiles = fix_index_name_after_groupby_quantile(
            groupby_except(peak_warming, "run_id").quantile(self.quantiles_of_interest),  # type: ignore # pandas-stubs confused
            new_name="quantile",
        )

        eoc_warming = set_new_single_value_levels(
            temperatures_in_line_with_assessment[2100], {"metric": 2100}
        )
        eoc_warming_quantiles = fix_index_name_after_groupby_quantile(
            groupby_except(eoc_warming, "run_id").quantile(self.quantiles_of_interest),  # type: ignore # pandas-stubs confused
            new_name="quantile",
        )
        peak_warming_year = set_new_single_value_levels(
            update_index_levels_func(
                temperatures_in_line_with_assessment.idxmax(axis="columns"),  # type: ignore # error in pandas-openscm
                {"unit": lambda x: "yr"},
            ),
            {"metric": "max_year"},
        )
        peak_warming_year_quantiles = fix_index_name_after_groupby_quantile(
            groupby_except(peak_warming_year, "run_id").quantile(
                self.quantiles_of_interest  # type: ignore # pandas-stubs out of date
            ),
            new_name="quantile",
        )

        exceedance_probabilities = get_exceedance_probabilities(
            temperatures_in_line_with_assessment,
            exceedance_thresholds_of_interest=self.exceedance_thresholds_of_interest,
            group_cols=["model", "scenario", "climate_model"],
            unit_col="unit",
            groupby_except_levels="run_id",
        )

        categories = categorise_scenarios(
            peak_warming_quantiles=peak_warming_quantiles,
            eoc_warming_quantiles=eoc_warming_quantiles,
            group_levels=["climate_model", "model", "scenario"],
            quantile_level="quantile",
        )

        timeseries_run_id = pd.concat([temperatures_in_line_with_assessment])
        timeseries_quantile = pd.concat(
            [temperatures_in_line_with_assessment_quantiles]
        )
        timeseries_exceedance_probabilities = pd.concat(
            [exceedance_probabilities_over_time]
        )

        metadata_run_id: pd.Series[float] = pd.concat(  # type: ignore # pandas-stubs out of date
            [peak_warming, eoc_warming, peak_warming_year]
        )
        metadata_quantile: pd.Series[float] = pd.concat(  # type: ignore # pandas-stubs out of date
            [
                peak_warming_quantiles,
                eoc_warming_quantiles,
                peak_warming_year_quantiles,
            ]
        )
        metadata_exceedance_probabilities = exceedance_probabilities
        metadata_categories = categories

        res = PostProcessingResult(
            timeseries_run_id=timeseries_run_id,
            timeseries_quantile=timeseries_quantile,
            timeseries_exceedance_probabilities=timeseries_exceedance_probabilities,
            metadata_run_id=metadata_run_id,
            metadata_quantile=metadata_quantile,
            metadata_exceedance_probabilities=metadata_exceedance_probabilities,
            metadata_categories=metadata_categories,
        )

        if self.run_checks:
            comparison_levels = ["model", "scenario", "climate_model"]
            for attr in [
                "timeseries_run_id",
                "timeseries_quantile",
                "timeseries_exceedance_probabilities",
                "metadata_run_id",
                "metadata_quantile",
                "metadata_exceedance_probabilities",
                "metadata_categories",
            ]:
                pd.testing.assert_index_equal(  # type: ignore # pandas-stubs out of date
                    getattr(res, attr)
                    .index.droplevel(
                        getattr(res, attr).index.names.difference(comparison_levels)
                    )
                    .drop_duplicates()
                    .reorder_levels(comparison_levels),
                    in_df.index.droplevel(
                        in_df.index.names.difference(comparison_levels)  # type: ignore # pandas-stubs out of date
                    )
                    .drop_duplicates()
                    .reorder_levels(comparison_levels),
                    check_order=False,
                )

        return res

    @classmethod
    def from_ar6_config(  # noqa: PLR0913
        cls,
        exceedance_thresholds_of_interest: tuple[float, ...] = tuple(
            np.arange(1.0, 4.01, 0.5)
        ),
        quantiles_of_interest: tuple[float, ...] = (
            0.05,
            0.10,
            1.0 / 6.0,
            0.33,
            0.50,
            0.67,
            5.0 / 6.0,
            0.90,
            0.95,
        ),
        raw_gsat_variable_in: str = "Surface Air Temperature Change",
        assessed_gsat_variable: str = "Surface Temperature (GSAT)",
        run_checks: bool = True,
        progress: bool = True,
        n_processes: int | None = multiprocessing.cpu_count(),
    ) -> AR6PostProcessor:
        """
        Initialise from the config used in AR6

        Parameters
        ----------
        exceedance_thresholds_of_interest
            The thresholds for which we are interested in exceedance probabilities

        quantiles_of_interest
            The quantiles we want to include in the results

        raw_gsat_variable_in
            Name of the variable that contains raw temperature output in the input

            The temperature output should be global-mean surface air temperature (GSAT).

        assessed_gsat_variable
            Name of the output variable that will contain temperature output

            This temperature output is in line with the
            (AR6) assessed historical warming.

        run_checks
            Should checks of the input and output data be performed?

            If this is turned off, things are faster,
            but error messages are much less clear if things go wrong.

        progress
            Should progress bars be shown for each operation?

        n_processes
            Number of processes to use for parallel processing.

            Set to 1 to process in serial.

        Returns
        -------
        :
            Initialised post-processor
        """
        if not all(q in quantiles_of_interest for q in [0.50, 0.33]):
            msg = (
                "quantiles_of_interest must contain 0.50 and 0.33 "
                "for the categorisation to work, "
                f"received {quantiles_of_interest=}"
            )
            raise AssertionError(msg)

        return cls(
            raw_gsat_variable_in=raw_gsat_variable_in,
            assessed_gsat_variable=assessed_gsat_variable,
            gsat_assessment_median=0.85,
            gsat_assessment_time_period=tuple(range(1995, 2014 + 1)),
            gsat_assessment_pre_industrial_period=tuple(range(1850, 1900 + 1)),
            quantiles_of_interest=quantiles_of_interest,
            exceedance_thresholds_of_interest=exceedance_thresholds_of_interest,
            run_checks=run_checks,
            n_processes=n_processes,
        )



"""
Useful assertions
"""

from __future__ import annotations

from collections.abc import Collection, Iterable
from typing import Any

import pandas as pd
from pandas.api.types import is_numeric_dtype


class DataIsNotAllNumericError(ValueError):
    """
    Raised when not all data in a [pd.DataFrame][pandas.DataFrame] is numeric
    """

    def __init__(self, df: pd.DataFrame, non_numeric_cols: Collection[Any]) -> None:
        """
        Initialise the error

        Parameters
        ----------
        df
            [pd.DataFrame][pandas.DataFrame] containing non-numeric data

        non_numeric_cols
            The columns that contain non-numeric data
        """
        # Including df in API, but not sure how to use it well right now
        # (not easy to just get the non-numeric values in a column,
        # because that's not a trivial question to ask [is "0" non-numeric or not?])
        error_msg = (
            f"The following columns contain non-numeric data: {non_numeric_cols}"
        )
        super().__init__(error_msg)


def assert_data_is_all_numeric(df: pd.DataFrame) -> None:
    """
    Assert that all data in a [pd.DataFrame][pandas.DataFrame] is numeric

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to check

    Raises
    ------
    DataIsNotAllNumericError
        If there are columns in `df` are not numeric
    """
    non_numeric = tuple(c for c in df if not is_numeric_dtype(df[c]))
    if non_numeric:
        raise DataIsNotAllNumericError(df=df, non_numeric_cols=non_numeric)


class MissingDataForTimesError(KeyError):
    """
    Raised when a [pd.DataFrame][pandas.DataFrame] is missing data for expected times
    """

    def __init__(
        self,
        df: pd.DataFrame,
        name: str,
        missing_times: Collection[Any],
        allow_nan: bool,
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        df
            [pd.DataFrame][pandas.DataFrame] that is missing expected index levels

        name
            Name of `df` to display in the error message

        missing_times
            Times in `df` that are missing data

        allow_nan
            Were NaN values allowed in the values of `times` when checking the data?
        """
        if allow_nan:
            error_msg = (
                f"{name} is missing data for the following times: "
                f"{missing_times}. "
                f"Available times: {df.columns}"
            )

        else:
            tmp = df[missing_times]
            nan_view = tmp[tmp.isnull().any(axis="columns")]
            error_msg = (
                f"{name} has NaNs for the following times: {missing_times}. "
                f"Rows with Nans:\n{nan_view}"
            )

        super().__init__(error_msg)


def assert_has_data_for_times(
    df: pd.DataFrame, name: str, times: Iterable[Any], allow_nan: bool
) -> None:
    """
    Assert that a [pd.DataFrame][pandas.DataFrame] has data for the given times

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to check

    name
        Name of `df` to display in the error message

    times
        Times (i.e. columns) that we expect to have data in `df`

    allow_nan
        Are NaN values allowed in the values of `times` (or should all data be non-Nan)?

    Raises
    ------
    MissingDataForTimesError
        The data in `df` does not contain all times in `times`.

        If `not allow_nan`, this will also be raised if any of the data in `df`
        contains NaN for a time in `times`.
    """
    missing_times = [v for v in times if v not in df.columns]
    if missing_times:
        raise MissingDataForTimesError(
            df=df,
            name=name,
            missing_times=missing_times,
            # Failed before we even considered NaN
            allow_nan=True,
        )

    if not allow_nan:
        nan_times = [v for v in times if df[v].isnull().any()]
        if nan_times:
            raise MissingDataForTimesError(
                df=df, name=name, missing_times=nan_times, allow_nan=allow_nan
            )


class MissingIndexLevelsError(KeyError):
    """
    Raised when a [pd.DataFrame][pandas.DataFrame] is missing expected index levels
    """

    def __init__(
        self,
        df: pd.DataFrame,
        missing_levels: Collection[Any],
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        df
            [pd.DataFrame][pandas.DataFrame] that is missing expected index levels

        missing_levels
            Levels that are missing from `df.index`
        """
        error_msg = (
            f"The DataFrame is missing the following index levels: {missing_levels}. "
            f"Available index levels: {df.index.names}"
        )
        super().__init__(error_msg)


def assert_has_index_levels(df: pd.DataFrame, levels: Iterable[Any]) -> None:
    """
    Assert that a [pd.DataFrame][pandas.DataFrame] has all the given levels in its index

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to check

    levels
        Levels that we expect to be in the index of `df`

    Raises
    ------
    MissingIndexLevelsError
        The index of `df` does not contain all levels in `levels`
    """
    missing_levels = [v for v in levels if v not in df.index.names]
    if missing_levels:
        raise MissingIndexLevelsError(df=df, missing_levels=missing_levels)


class IndexIsNotMultiIndexError(TypeError):
    """
    Raised when the index is not a [pd.MultiIndex][pandas.MultiIndex]
    """

    def __init__(self, df: pd.DataFrame) -> None:
        """
        Initialise the error

        Parameters
        ----------
        df
            [pd.DataFrame][pandas.DataFrame]
        """
        error_msg = (
            f"The index is not a `pd.MultiIndex`, instead we have {type(df.index)=}"
        )
        super().__init__(error_msg)


def assert_index_is_multiindex(df: pd.DataFrame) -> None:
    """
    Assert that the index is a [pd.MultiIndex][pandas.MultiIndex]

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to check

    Raises
    ------
    IndexIsNotMultiIndexError
        The index of `df` is not a [pd.MultiIndex][pandas.MultiIndex]
    """
    if not isinstance(df.index, pd.MultiIndex):
        raise IndexIsNotMultiIndexError(df)


class NotAllowedMetadataValuesError(ValueError):
    """
    Raised when a [pd.DataFrame][pandas.DataFrame] contains disallowed metadata values
    """

    def __init__(
        self,
        df: pd.DataFrame,
        metadata_key: Any,
        disallowed_values: Collection[Any],
        allowed_values: Collection[Any],
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        df
            [pd.DataFrame][pandas.DataFrame] that contains diasallowed metadata values

        metadata_key
            The metadata key which is being considered (e.g. "variable", "unit")

        disallowed_values
            The values which are not allowed but appear in `df`

        allowed_values
            The values which are allowed for `metadata_key`
        """
        error_msg = (
            f"The DataFrame contains disallowed values for {metadata_key}: "
            f"{disallowed_values}. "
            f"Allowed values: {allowed_values}"
        )
        super().__init__(error_msg)


def assert_metadata_values_all_allowed(
    df: pd.DataFrame, metadata_key: Any, allowed_values: Collection[Any]
) -> None:
    """
    Assert that a [pd.DataFrame][pandas.DataFrame] only contains allowed metadata values

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to check

    metadata_key
        The metadata key to check (e.g. "variable", "unit")

    allowed_values
        The values which are allowed for this metadata key

    Raises
    ------
    NotAllowedMetadataValuesError
        There is metadata for `metadata_key` in `df` that is not in `allowed_values`.
    """
    disallowed_values = [
        v
        for v in df.index.get_level_values(metadata_key).unique()
        if v not in allowed_values
    ]
    if disallowed_values:
        raise NotAllowedMetadataValuesError(
            df=df,
            metadata_key=metadata_key,
            disallowed_values=disallowed_values,
            allowed_values=allowed_values,
        )


def assert_only_working_on_variable_unit_variations(indf: pd.DataFrame) -> None:
    """
    Assert that we're only working on variations in variable and unit

    In other words, we don't have variations in scenarios, models etc.

    Parameters
    ----------
    indf
        Data to verify

    Raises
    ------
    AssertionError
        There are variations in columns other than variable and unit
    """
    variations_in_other_cols = indf.index.droplevel(["variable", "unit"]).unique()
    if len(variations_in_other_cols) > 1:
        msg = f"variations_in_other_cols=\n{variations_in_other_cols}"
        raise AssertionError(msg)


def assert_only_working_on_variable_unit_region_variations(indf: pd.DataFrame) -> None:
    """
    Assert that we're only working on variations in variable, unit and region

    In other words, we don't have variations in scenarios, models etc.

    Parameters
    ----------
    indf
        Data to verify

    Raises
    ------
    AssertionError
        There are variations in columns other than variable and unit
    """
    variations_in_other_cols = indf.index.droplevel(
        ["variable", "unit", "region"]
    ).unique()
    if len(variations_in_other_cols) > 1:
        msg = f"variations_in_other_cols=\n{variations_in_other_cols}"
        raise AssertionError(msg)



"""
Pre-processing part of the workflow
"""

from __future__ import annotations

import multiprocessing
from collections.abc import Mapping
from functools import partial
from typing import TYPE_CHECKING, Callable

import pandas as pd
from attrs import define
from pandas_openscm.index_manipulation import update_index_levels_func
from pandas_openscm.parallelisation import ParallelOpConfig, apply_op_parallel_progress

from gcages.assertions import (
    assert_data_is_all_numeric,
    assert_has_data_for_times,
    assert_has_index_levels,
    assert_index_is_multiindex,
    assert_only_working_on_variable_unit_variations,
)
from gcages.exceptions import MissingOptionalDependencyError
from gcages.renaming import SupportedNamingConventions, convert_variable_name
from gcages.units_helpers import strip_pint_incompatible_characters_from_units

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec

    P = ParamSpec("P")


# As a note, if you wanted to re-use these helpers somewhere else,
# you would want to test them carefully first,
# it's really not clear to me that they are well-written or obvious.
def add_conditional_sums(
    indf: pd.DataFrame,
    conditional_sums: tuple[tuple[str, tuple[str, ...]], ...],
    copy_on_entry: bool = True,
) -> pd.DataFrame:
    """
    Add sums to a [pd.DataFrame][pandas.DataFrame] if all components are present

    Parameters
    ----------
    indf
        Data to add sums to

    conditional_sums
        Definition of the conditional sums.

        The first element of each sub-tuple is the name of the variable to add.
        The second element are its components.
        If the variable is added, all the sub-components are dropped.
        All components must be present for the variable to be added.
        If the variable is already there, the sum is not re-calculated or checked.

    copy_on_entry
        Should the data be copied on entry?

    Returns
    -------
    :
        `indf` with conditional sums added if all enabling conditions were fulfilled.
    """
    try:
        from pandas_indexing.core import concat
        from pandas_indexing.selectors import isin
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "add_conditional_sums", requirement="pandas_indexing"
        ) from exc

    assert_only_working_on_variable_unit_variations(indf)

    if copy_on_entry:
        out = indf.copy()

    else:
        out = indf

    for v_target, v_sources in conditional_sums:
        existing_vars: pd.MultiIndex = out.pix.unique("variable")  # type: ignore
        if v_target not in existing_vars:
            if all(v in existing_vars for v in v_sources):
                locator_sources = isin(variable=v_sources)
                to_add = out.loc[locator_sources]

                tmp = to_add.groupby(list(set(to_add.index.names) - {"variable"})).sum(
                    min_count=len(v_sources)
                )
                tmp = tmp.pix.assign(variable=v_target)
                out = concat([out.loc[~locator_sources], tmp], axis="index")

    return out


def reclassify_variables(
    indf: pd.DataFrame,
    reclassifications: Mapping[str, tuple[str, ...]],
    copy_on_entry: bool = True,
) -> pd.DataFrame:
    """
    Reclassify variables

    Parameters
    ----------
    indf
        Data to add sums to

    reclassifications
        Definition of the reclassifications.

        For each variable (key) in `reclassifications`, the variables in its value
        will be reclassified as part of its total.

        For example, if `reclassifications` is

        ```python
        {"var_a": ("var_b", "var_c")}
        ```

        then if "var_b" or "var_c" (or both) is in `indf`,
        they will be removed and their contents will be added to the total of `var_a`.

    copy_on_entry
        Should the data be copied on entry?

    Returns
    -------
    :
        `indf`, reclassified as needed.
    """
    try:
        from pandas_indexing.selectors import isin
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "reclassify_variables", requirement="pandas_indexing"
        ) from exc

    assert_only_working_on_variable_unit_variations(indf)

    if copy_on_entry:
        out = indf.copy()

    else:
        out = indf

    for v_target, v_sources in reclassifications.items():
        locator_sources = isin(variable=v_sources)
        to_add = out.loc[locator_sources]
        if not to_add.empty:
            out.loc[isin(variable=v_target)] += to_add.sum()  # type: ignore
            out = out.loc[~locator_sources]

    return out


def condtionally_remove_variables(
    indf: pd.DataFrame,
    conditional_removals: tuple[tuple[str, tuple[str, ...]], ...],
    copy_on_entry: bool = True,
) -> pd.DataFrame:
    """
    Conditionally remove variables

    Parameters
    ----------
    indf
        Data to add sums to

    conditional_removals
        Definition of the conditional removals.

        For each tuple, the first element defines the variable that can be removed.
        This variable will be removed if all variables in the tuple's second element
        are present in `indf`.

    copy_on_entry
        Should the data be copied on entry?

    Returns
    -------
    :
        `indf` with variables removed according to this function's logic.
    """
    try:
        from pandas_indexing.selectors import isin
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "condtionally_remove_variables", requirement="pandas_indexing"
        ) from exc

    assert_only_working_on_variable_unit_variations(indf)

    if copy_on_entry:
        out = indf.copy()

    else:
        out = indf

    for v_drop, v_sub_components in conditional_removals:
        existing_vars: pd.MultiIndex = out.pix.unique("variable")  # type: ignore
        if v_drop in existing_vars and all(
            v in existing_vars for v in v_sub_components
        ):
            out = out.loc[~isin(variable=v_drop)]

    return out


def drop_variables_if_identical(
    indf: pd.DataFrame,
    drop_if_identical: tuple[tuple[str, str], ...],
    copy_on_entry: bool = True,
) -> pd.DataFrame:
    """
    Drop variables if they are identical to another variable

    Parameters
    ----------
    indf
        Data to add sums to

    drop_if_identical
        Definition of the variables that can be dropped.

        For each tuple, the first element defines the variable that can be removed
        and the second element defines the variable to compare it to.
        If the variable to drop has the same values as the variable to compare to,
        it is dropped.

    copy_on_entry
        Should the data be copied on entry?

    Returns
    -------
    :
        `indf` with variables removed according to this function's logic.
    """
    try:
        from pandas_indexing.selectors import isin
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "drop_variables_if_identical", requirement="pandas_indexing"
        ) from exc

    assert_only_working_on_variable_unit_variations(indf)

    if copy_on_entry:
        out = indf.copy()

    else:
        out = indf

    for v_drop, v_check in drop_if_identical:
        existing_vars: pd.MultiIndex = out.pix.unique("variable")  # type: ignore
        if all(v in existing_vars for v in (v_drop, v_check)):
            # Should really use isclose here, but we didn't in AR6
            # so we get some funny reporting for weird scenarios
            # e.g. C3IAM 2.0 2C-hybrid
            if (
                (
                    out.loc[isin(variable=v_drop)]
                    .reset_index("variable", drop=True)
                    .dropna(axis="columns")
                    == out.loc[isin(variable=v_check)]
                    .reset_index("variable", drop=True)
                    .dropna(axis="columns")
                )
                .all()
                .all()
            ):
                out = out.loc[~isin(variable=v_drop)]

    return out


def run_parallel_pre_processing(  # noqa: PLR0913
    indf: pd.DataFrame,
    func_to_call: Callable[Concatenate[pd.DataFrame, P], pd.DataFrame],
    groups: tuple[str, ...] = ("model", "scenario"),
    progress: bool = True,
    progress_bar_desc: str | None = None,
    n_processes: int | None = multiprocessing.cpu_count(),
    *args: P.args,
    **kwargs: P.kwargs,
) -> pd.DataFrame:
    """
    Run a pre-processing step in parallel

    Parameters
    ----------
    indf
        Input data to process

    func_to_call
        Function to apply to each group in `indf`

    groups
        Columns to use to group the data in `indf`

    progress
        Should a progress bar be displayed?

    progress_bar_desc
        If `progress`, the description of the progress bar.

        If not supplied, we use a default description.

    n_processes
        Number of parallel processes to use

    **kwargs
        Passed to `run_parallel`

    Returns
    -------
    :
        Result of calling `func_to_call` on each group in `indf`.
    """
    if progress and progress_bar_desc is None:
        progress_bar_desc = f"{', '.join(groups)} combinations"

    res = pd.concat(
        apply_op_parallel_progress(
            func_to_call,
            (gdf for _, gdf in indf.groupby(list(groups))),
            ParallelOpConfig.from_user_facing(
                progress=progress,
                progress_results_kwargs=dict(desc=progress_bar_desc),
                max_workers=n_processes,
            ),
            *args,
            **kwargs,
        )
    )

    return res


@define
class AR6PreProcessor:
    """
    Pre-processor that follows the same logic as was used in AR6

    If you want exactly the same behaviour as in AR6,
    initialise using [`from_ar6_config`][(c)]
    """

    emissions_out: tuple[str, ...]
    """
    Names of emissions that can be included in the result of pre-processing

    Not all these emissions need to be there,
    but any names which are not in this list will be removed as part of pre-processing.
    """

    negative_value_not_small_threshold: float
    """
    Threshold which defines when a negative value is not small

    Non-CO2 emissions less than this that are negative
    are not automatically set to zero.
    """

    conditional_sums: tuple[tuple[str, tuple[str, ...]], ...] | None = None
    """
    Specification for variables that can be created from other variables

    Form:

    ```python
    (
        (variable_that_can_be_created, (component_1, component_2)),
        ...
    )
    ```

    The variable that can be created is only created
    if all the variables it depends on are present.
    """

    reclassifications: Mapping[str, tuple[str, ...]] | None = None
    """
    Variables that should be reclassified as being part of another variable

    Form:

    ```python
    {
        variable_to_add_to: (variable_to_rename_1, variable_to_rename_2),
        ...
    }
    ```

    For example
    ```python
    {
        "Emissions|CO2|Energy and Industrial Processes": (
            "Emissions|CO2|Other",
            "Emissions|CO2|Waste",
        )
    }
    ```
    """

    conditional_removals: tuple[tuple[str, tuple[str, ...]], ...] | None = None
    """
    Specification for variables that can be removed if other variables are present

    Form:

    ```python
    (
        (variable_that_can_be_removed, (component_1, component_2)),
        ...
    )
    ```

    The variable that can be removed is only removed
    if all the variables it depends on are present.
    """

    drop_if_identical: tuple[tuple[str, str], ...] | None = None
    """
    Variables that can be dropped if they are idential to another variable

    Form:

    ```python
    (
        (variable_that_can_be_removed, variable_to_compare_to),
        ...
    )
    ```

    The variable that can be removed is only removed
    if its values are identical to the variable it is compared to.
    """

    run_checks: bool = True
    """
    If `True`, run checks on both input and output data

    If you are sure about your workflow,
    you can disable the checks to speed things up
    (but we don't recommend this unless you really
    are confident about what you're doing).
    """

    progress: bool = True
    """
    Should progress bars be shown for each operation?
    """

    n_processes: int | None = multiprocessing.cpu_count()
    """
    Number of processes to use for parallel processing.

    Set to `None` to process in serial.
    """

    def __call__(self, in_emissions: pd.DataFrame) -> pd.DataFrame:
        """
        Pre-process

        Parameters
        ----------
        in_emissions
            Emissions to pre-process

        Returns
        -------
        :
            Pre-processed emissions
        """
        try:
            from pandas_indexing.selectors import isin, ismatch
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "AR6PreProcessor.__call__", requirement="pandas_indexing"
            ) from exc

        if self.run_checks:
            assert_index_is_multiindex(in_emissions)
            assert_data_is_all_numeric(in_emissions)
            assert_has_index_levels(in_emissions, ["variable", "unit"])

        # Remove any rows with only zero (custom AR6 thing)
        in_emissions = in_emissions[
            ~(((in_emissions == 0.0) | in_emissions.isnull()).all(axis="columns"))
        ]

        rp = partial(
            run_parallel_pre_processing,
            progress=self.progress,
            n_processes=self.n_processes,
        )
        if self.conditional_sums is not None:
            in_emissions = rp(  # type: ignore
                in_emissions,
                func_to_call=add_conditional_sums,
                progress_bar_desc=(
                    "For each model-scenario, calculating conditional sums"
                ),
                conditional_sums=self.conditional_sums,
            )

        if self.reclassifications is not None:
            in_emissions = rp(  # type: ignore
                in_emissions,
                func_to_call=reclassify_variables,
                progress_bar_desc="For each model-scenario, reclassifying variables",
                reclassifications=self.reclassifications,
            )

        if self.conditional_removals is not None:
            in_emissions = rp(  # type: ignore
                in_emissions,
                func_to_call=condtionally_remove_variables,
                progress_bar_desc=(
                    "For each model-scenario, conditionally removing variables"
                ),
                conditional_removals=self.conditional_removals,
            )

        if self.drop_if_identical is not None:
            in_emissions = rp(  # type: ignore
                in_emissions,
                func_to_call=drop_variables_if_identical,
                progress_bar_desc=(
                    "For each model-scenario, dropping variables if they are identical"
                ),
                drop_if_identical=self.drop_if_identical,
            )

        # Negative value handling
        co2_locator = ismatch(variable="**CO2**")
        in_emissions.loc[~co2_locator] = in_emissions.loc[~co2_locator].where(
            # Where these conditions are true, keep the original data.
            (in_emissions.loc[~co2_locator] > 0)
            | (in_emissions.loc[~co2_locator] < self.negative_value_not_small_threshold)
            | in_emissions.loc[~co2_locator].isnull(),
            # Otherwise, set to zero
            other=0.0,
        )

        res: pd.DataFrame = in_emissions.loc[isin(variable=self.emissions_out)]

        # Strip out any units that won't play nice with pint
        res = strip_pint_incompatible_characters_from_units(
            res, units_index_level="unit"
        )

        # Convert to gcages naming conventions
        res = update_index_levels_func(
            res,
            {
                "variable": partial(
                    convert_variable_name,
                    from_convention=SupportedNamingConventions.IAMC,
                    to_convention=SupportedNamingConventions.GCAGES,
                )
            },
        )

        if self.run_checks:
            # AR6 required emissions for these years after pre-processing,
            # for some reason
            required_years = list(range(2020, 2100 + 1, 10))
            assert_has_data_for_times(
                res, name="res", times=required_years, allow_nan=False
            )

        return res

    @classmethod
    def from_ar6_config(
        cls,
        run_checks: bool = True,
        progress: bool = True,
        n_processes: int | None = multiprocessing.cpu_count(),
    ) -> AR6PreProcessor:
        """
        Initialise from config that was used in AR6

        Parameters
        ----------
        run_checks
            Should checks of the input and output data be performed?

            If this is turned off, things are faster,
            but error messages are much less clear if things go wrong.

        progress
            Should a progress bar be shown for each operation?

        n_processes
            Number of processes to use for parallel processing.

            Set to `None` to process in serial.

        Returns
        -------
        :
            Initialised Pre-processor
        """
        ar6_emissions_for_harmonisation_iamc = tuple(
            v
            for v in (
                "Emissions|BC",
                "Emissions|PFC|C2F6",
                "Emissions|PFC|C6F14",
                "Emissions|PFC|CF4",
                "Emissions|CO",
                "Emissions|CO2",
                "Emissions|CO2|AFOLU",
                "Emissions|CO2|Energy and Industrial Processes",
                "Emissions|CH4",
                # "Emissions|F-Gases",  # Not used
                # "Emissions|HFC",  # Not used
                "Emissions|HFC|HFC125",
                "Emissions|HFC|HFC134a",
                "Emissions|HFC|HFC143a",
                "Emissions|HFC|HFC227ea",
                "Emissions|HFC|HFC23",
                # 'Emissions|HFC|HFC245ca',  # all nan in historical dataset (RCMIP)
                # "Emissions|HFC|HFC245fa",  # not in historical dataset (RCMIP)
                "Emissions|HFC|HFC32",
                "Emissions|HFC|HFC43-10",
                "Emissions|N2O",
                "Emissions|NH3",
                "Emissions|NOx",
                "Emissions|OC",
                # "Emissions|PFC",  # Not used
                "Emissions|SF6",
                "Emissions|Sulfur",
                "Emissions|VOC",
            )
        )
        conditional_sums = (
            (
                "Emissions|CO2|Energy and Industrial Processes",
                (
                    "Emissions|CO2|Industrial Processes",
                    "Emissions|CO2|Energy",
                ),
            ),
        )
        reclassifications = {
            "Emissions|CO2|Energy and Industrial Processes": (
                "Emissions|CO2|Other",
                "Emissions|CO2|Waste",
            )
        }
        conditional_removals = (
            (
                "Emissions|CO2",
                (
                    "Emissions|CO2|Energy and Industrial Processes",
                    "Emissions|CO2|AFOLU",
                ),
            ),
        )
        drop_if_identical = (
            ("Emissions|CO2", "Emissions|CO2|Energy and Industrial Processes"),
            ("Emissions|CO2", "Emissions|CO2|AFOLU"),
        )

        return cls(
            emissions_out=ar6_emissions_for_harmonisation_iamc,
            negative_value_not_small_threshold=-0.1,
            conditional_sums=conditional_sums,
            reclassifications=reclassifications,
            conditional_removals=conditional_removals,
            drop_if_identical=drop_if_identical,
            run_checks=run_checks,
            n_processes=n_processes,
            progress=progress,
        )



"""
Aggregation helpers
"""

from __future__ import annotations

import string
from functools import partial
from typing import Callable

import pandas as pd
from pandas_openscm.grouping import groupby_except
from pandas_openscm.indexing import multi_index_lookup, multi_index_match

import gcages.index_manipulation
from gcages.exceptions import MissingOptionalDependencyError
from gcages.index_manipulation import combine_species, set_new_single_value_levels
from gcages.testing import compare_close


def aggregate_df_level(
    indf: pd.DataFrame,
    level: str,
    on_clash: str = "raise",
    component_separator: str = "|",
    min_components_output: int = 1,
) -> pd.DataFrame:
    """
    Aggregate a level in a [pd.DataFrame][pandas.DataFrame]

    Here, aggregate means 'walk up the components in the level values'
    and create their totals.
    For example, if `indf` has a metadata value like
    "Emission|CO2|Energy|Demand"
    then this could walk up the tree to create
    "Emission|CO2|Energy" and "Emissions|CO2"
    metadata values too.

    Parameters
    ----------
    indf
        Data to process

    level
        Level to aggregate

    on_clash
        What to do if there is a clash while aggregating.

        Options:

        - "raise": raise an error
        - "verify": verify that the aggregated data
                    is consistent with the existing data,
                    raise an error if not
        - "overwrite": overwrite the existing data with the aggregated data

    component_separator
        Separator between components within the values of `level`

    min_components_output
        Minimum number of components to include in the output

        This helps avoid creating aggregates for components you don't care about
        (e.g. you might not care about a "Emissions" aggregate
        if you have metadata values like "Emissions|CO2" and "Emissions|CH4").

    Returns
    -------
    :
        Aggregated data (i.e. both the input and the newly aggregated timeseries)
    """
    # TODO: move this into pandas-openscm
    # will require writing our own extract and format functions
    # will require moving compare_close too
    try:
        from pandas_indexing.core import extractlevel, formatlevel
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "assert_frame_equal", requirement="pandas_indexing"
        ) from exc
    level_groups = {
        n_levels: df
        for n_levels, df in indf.groupby(
            indf.index.get_level_values(level).str.count(rf"\{component_separator}")
        )
    }
    levels_r = range(min_components_output, max(level_groups) + 1)[::-1]
    # Start by storing the bottom level
    res_d = {levels_r[0]: level_groups[levels_r[0]]}
    for n_levels in levels_r[1:]:
        # Get everything we have already handled
        # at the level below the level of interest
        to_aggregate = res_d[n_levels + 1]

        # Aggregate
        to_aggregate.index = to_aggregate.index.remove_unused_levels()  # type: ignore # pandas-stubs confused

        level_splits = [
            f"{level}_{string.ascii_lowercase[i]}" for i in range(n_levels + 1 + 1)
        ]
        extract_str = component_separator.join(["{" + ls + "}" for ls in level_splits])
        to_aggregate_split = extractlevel(to_aggregate, **{level: extract_str})

        to_aggregate_sum = groupby_except(to_aggregate_split, level_splits[-1]).sum()

        to_aggregate_sum_combined = formatlevel(
            to_aggregate_sum,
            **{
                level: component_separator.join(
                    ["{" + ls + "}" for ls in level_splits[:-1]]
                )
            },
            drop=True,
        )

        keep_at_level = [to_aggregate_sum_combined]
        if n_levels in level_groups:
            # Check if any of the aggregations clash with the input
            indf_at_aggregated_level = level_groups[n_levels]
            clash_locator = multi_index_match(
                indf_at_aggregated_level.index,  # type: ignore # pandas-stubs confused
                to_aggregate_sum_combined.index,
            )
            if not clash_locator.any():
                # No clashing data so simply keep all of `indf_at_aggregated_level`
                keep_at_level.append(indf_at_aggregated_level)

            elif on_clash == "raise":
                clash = indf_at_aggregated_level[clash_locator]
                msg = f"Reaggregated levels are in the input. Clashing levels: {clash}"
                # TODO: switch to custom error
                raise ValueError(msg)

            elif on_clash == "verify":
                indf_compare = indf_at_aggregated_level[clash_locator]
                to_aggregate_sum_combined_compare = multi_index_lookup(
                    to_aggregate_sum_combined,
                    indf_compare.index.remove_unused_levels(),  # type: ignore # pandas-stubs confused
                )
                comparison = compare_close(
                    left=indf_compare,
                    right=to_aggregate_sum_combined_compare,
                    left_name="indf",
                    right_name="aggregated_sum",
                    # **tolerances,
                )
                if not comparison.empty:
                    raise NotImplementedError

            elif on_clash == "overwrite":
                not_clashing = indf_at_aggregated_level[~clash_locator]
                if not_clashing.empty:
                    # (Nothing to keep from input so do nothing here)
                    pass

                else:
                    # Only keep what doesn't clash, effectively overwriting the rest
                    # by using to_aggregate_sum_combined
                    # instead
                    keep_at_level.append(not_clashing)

            else:
                raise NotImplementedError(on_clash)

        res_d[n_levels] = pd.concat(
            [
                df.reorder_levels(to_aggregate_sum_combined.index.names)
                for df in keep_at_level
            ]
        )

    res = pd.concat(
        [
            df.reorder_levels(res_d[min_components_output].index.names)
            for df in res_d.values()
        ]
    )

    return res


def get_region_sector_sum(
    indf: pd.DataFrame,
    region_level: str = "region",
    world_region: str = "World",
    split_sectors: Callable[[pd.DataFrame], pd.DataFrame] = partial(
        gcages.index_manipulation.split_sectors, bottom_level="sectors"
    ),
    sectors_level: str = "sectors",
) -> pd.DataFrame:
    """
    Get the sum over regions and sectors

    Parameters
    ----------
    indf
        Input data to sum

    region_level
        Region level in the data index

    world_region
        The value used when the data represents the sum over all regions

    split_sectors
        Callable to use to split sectors from other levels in `indf`'s index

    sectors_level
        Name of the sectors level once the data is split

        (Should be consistent with `split_sectors`)

    Returns
    -------
    :
        Region-sector sum of `indf`

        To meet other conventions, the output has a region level
        with value `world_region` (even though this is a super weird convention).
    """
    # Blind sum so double counting is possible if you're not careful
    # Can't avoid double counting as that is defined by data model,
    # not by "|" separation rules
    # Implicitly assumes that `split_sectors` will work
    res = set_new_single_value_levels(
        combine_species(
            groupby_except(split_sectors(indf), [region_level, sectors_level]).sum()
        ),
        {region_level: world_region},
    )

    return res



"""
General post-processing tools
"""

from __future__ import annotations

import pandas as pd
from attrs import define


@define
class PostProcessingResult:
    """
    Results of post-processing

    The data is separated into tables that have the same index levels.
    This simplifies further processing
    (at the expense of making it more complex to get data into the object).
    """

    timeseries_run_id: pd.DataFrame
    """Timeseries that includes a run_id index level"""

    timeseries_quantile: pd.DataFrame
    """Timeseries that includes a quantile index level"""

    timeseries_exceedance_probabilities: pd.DataFrame
    """
    Timeseries of exceedance probabilities

    These are reported separately because they contain
    extra index levels to handle the threshold information.
    """

    metadata_run_id: pd.Series[float]
    """Metadata that includes a run_id index level"""

    metadata_quantile: pd.Series[float]
    """Metadata that includes a quantile index level"""

    metadata_exceedance_probabilities: pd.Series[float]
    """
    Metadata of exceedance probabilities

    These are reported separately because they contain
    extra index levels to handle the threshold information.
    """

    metadata_categories: pd.Series[str]
    """
    Metadata of categories
    """



"""
Reaggregation of timeseries from raw reporting to sectors needed for gridding

The idea here is that we receive raw data following some variable specification
Based on this, we reaggregate to the variables needed for gridding
(see [gcages.cmip7_scenariomip.gridding_emissions][]).
In order to do the reaggregation sensibly,
two things must be true:

1. all the timeseries we require must be there
1. the data must be internally consistent
    - including consideration of any optional timeseries

Reaggregation is a data problem
i.e. the hard part is making sure
that the data we receive matches our data model.
As a result, the code is highly coupled with the data we expect
(writing general solutions is hard).
This is why we have written the code
that supports each data model in a standalone module,
rather than trying to write a general solution
(which was extremely difficult when we tried to do it that way from the start,
we think because it creates couplings
which are incredibly difficult to reason through).
"""

from __future__ import annotations

from gcages.cmip7_scenariomip.pre_processing.reaggregation.basic import (
    ReaggregatorBasic,
)
from gcages.cmip7_scenariomip.pre_processing.reaggregation.common import (
    ToCompleteResult,
)

__all__ = ["ReaggregatorBasic", "ToCompleteResult"]



"""
Harmonisation using [aneris](https://aneris.readthedocs.io/)
"""

from __future__ import annotations

import multiprocessing
from typing import Any

import attr
import pandas as pd
from attrs import define, field
from pandas_openscm.parallelisation import ParallelOpConfig, apply_op_parallel_progress

import gcages.aneris_helpers
from gcages.assertions import (
    NotAllowedMetadataValuesError,
    assert_data_is_all_numeric,
    assert_has_data_for_times,
    assert_has_index_levels,
    assert_index_is_multiindex,
    assert_metadata_values_all_allowed,
)
from gcages.harmonisation.common import assert_harmonised


@define
class AnerisHarmoniser:
    """
    Harmoniser that uses [aneris](https://aneris.readthedocs.io/)
    """

    historical_emissions: pd.DataFrame = field()
    """
    Historical emissions to use for harmonisation
    """

    harmonisation_year: int
    """
    Year in which to harmonise
    """

    aneris_overrides: pd.Series[str] | None = field(default=None)
    """
    Overrides to supply to `aneris.convenience.harmonise_all`

    For source code and docs,
    see e.g. [https://github.com/iiasa/aneris/blob/v0.4.2/src/aneris/convenience.py]().
    """

    run_checks: bool = True
    """
    If `True`, run checks on both input and output data

    If you are sure about your workflow,
    you can disable the checks to speed things up
    (but we don't recommend this unless you really
    are confident about what you're doing).
    """

    variable_level: str = "variable"
    """
    Level in data indexes that represents the variable of the timeseries
    """

    region_level: str = "region"
    """
    Level in data indexes that represents the region of the timeseries
    """

    unit_level: str = "unit"
    """
    Level in data indexes that represents the unit of the timeseries
    """

    scenario_group_levels: list[str] = field(factory=lambda: ["model", "scenario"])
    """
    Levels in data indexes to use to group data into scenarios

    Here, 'scenarios' means groups of timeseries
    that will be run through a climate model.
    """

    progress: bool = True
    """
    Should progress bars be shown for each operation?
    """

    n_processes: int | None = multiprocessing.cpu_count()
    """
    Number of processes to use for parallel processing.

    Set to `None` to process in serial.
    """

    @aneris_overrides.validator
    def validate_aneris_overrides(
        self, attribute: attr.Attribute[Any], value: pd.Series[str] | None
    ) -> None:
        """
        Validate the aneris overrides value

        If `self.run_checks` is `False`, then this is a no-op
        """
        if value is None:
            return

        if not self.run_checks:
            return

    @historical_emissions.validator
    def validate_historical_emissions(
        self, attribute: attr.Attribute[Any], value: pd.DataFrame
    ) -> None:
        """
        Validate the historical emissions value

        If `self.run_checks` is `False`, then this is a no-op
        """
        if not self.run_checks:
            return

        assert_index_is_multiindex(value)
        assert_data_is_all_numeric(value)
        assert_has_index_levels(
            value, [self.variable_level, self.region_level, self.unit_level]
        )
        assert_has_data_for_times(
            value,
            name="historical_emissions",
            times=[self.harmonisation_year],
            allow_nan=False,
        )

    def __call__(self, in_emissions: pd.DataFrame) -> pd.DataFrame:
        """
        Harmonise

        Parameters
        ----------
        in_emissions
            Emissions to harmonise

        Returns
        -------
        :
            Harmonised emissions
        """
        if self.run_checks:
            assert_index_is_multiindex(in_emissions)
            assert_data_is_all_numeric(in_emissions)
            assert_has_index_levels(
                in_emissions,
                [
                    self.variable_level,
                    self.region_level,
                    self.unit_level,
                    # Needed for parallelisation
                    *self.scenario_group_levels,
                ],
            )
            assert_has_data_for_times(
                in_emissions,
                name="in_emissions",
                times=[self.harmonisation_year],
                allow_nan=False,
            )

            try:
                assert_metadata_values_all_allowed(
                    in_emissions,
                    metadata_key=self.variable_level,
                    allowed_values=self.historical_emissions.index.get_level_values(
                        self.variable_level
                    ).unique(),
                )
            except NotAllowedMetadataValuesError as exc:
                msg = "The input emissions contains values that aren't in history"
                raise ValueError(msg) from exc

        harmonised_df = pd.concat(
            apply_op_parallel_progress(
                func_to_call=gcages.aneris_helpers.harmonise_scenario,
                iterable_input=(
                    gdf for _, gdf in in_emissions.groupby(self.scenario_group_levels)
                ),
                parallel_op_config=ParallelOpConfig.from_user_facing(
                    progress=self.progress,
                    max_workers=self.n_processes,
                    progress_results_kwargs=dict(desc="Scenarios to harmonise"),
                ),
                history=self.historical_emissions,
                year=self.harmonisation_year,
                overrides=self.aneris_overrides,
            )
        )

        if self.run_checks:
            assert_harmonised(
                harmonised_df,
                history=self.historical_emissions,
                harmonisation_time=self.harmonisation_year,
            )

            pd.testing.assert_index_equal(
                harmonised_df.index,
                in_emissions.index,
                check_order=False,  # type: ignore # pandas-stubs out of date
            )
            if harmonised_df.columns.dtype != in_emissions.columns.dtype:
                msg = (
                    "Column type has changed: "
                    f"{harmonised_df.columns.dtype=} {in_emissions.columns.dtype=}"
                )
                raise AssertionError(msg)

        return harmonised_df
